# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, you can obtain one at http://mozilla.org/MPL/2.0/.

import json
import os.path

from marionette_harness import MarionetteTestCase

here = os.path.dirname(__file__)

with open(os.path.join(here, "scripts", "tabs.js")) as script:
    get_tabs = script.read()
mail_3pane_tab = {"mode": "mail3PaneTab"}
close_tabs = """
    const tabmail = document.getElementById("tabmail");
    tabmail.closeOtherTabs(0);
"""


class TestSessionStore(MarionetteTestCase):
    def check_json_file(self, expected):
        """Checks the session store JSON file has the right info."""
        json_path = os.path.join(self.marionette.profile_path, "session.json")
        with open(json_path, "r") as json_file:
            actual = json.load(json_file)

        def assert_similar(actual_value, expected_value):
            """Checks that `actual_value` matches `expected_value` by type and content."""
            if isinstance(expected_value, list):
                # This is a strict check: the items in `expected_value` must be in `actual_value`,
                # and in the same order.
                self.assertIsInstance(actual_value, list)
                self.assertEqual(len(actual_value), len(expected_value))
                for i in range(0, len(expected_value)):
                    assert_similar(actual_value[i], expected_value[i])
                return

            if isinstance(expected_value, dict):
                # This is a lenient check: if a key isn't in `expected_value`, it isn't checked.
                self.assertIsInstance(actual_value, dict)
                for k in expected_value.keys():
                    assert_similar(actual_value[k], expected_value[k])
                return

            self.assertEqual(actual_value, expected_value)

        assert_similar(actual["windows"][0], expected[0])

    def test_only_3pane_tab(self):
        self.marionette.set_context(self.marionette.CONTEXT_CHROME)
        self.assertEqual("mail:3pane", self.marionette.get_window_type())
        self.marionette.execute_script(
            """
            MailServices.accounts.createLocalMailAccount();
            """
        )
        self.marionette.quit(in_app=True)

        self.check_json_file(
            [
                {
                    "type": "3pane",
                    "tabs": {
                        "tabs": [
                            {
                                "mode": "mail3PaneTab",
                                "state": {
                                    "firstTab": True,
                                    "folderPaneVisible": True,
                                    "folderURI": "mailbox://nobody@Local%20Folders",
                                    "messagePaneVisible": True,
                                },
                            }
                        ]
                    },
                }
            ]
        )

        self.marionette.start_session()
        self.marionette.set_context(self.marionette.CONTEXT_CHROME)
        self.assertEqual("mail:3pane", self.marionette.get_window_type())
        tabs = self.marionette.execute_async_script(get_tabs)
        self.assertEqual([mail_3pane_tab], tabs)
        self.marionette.execute_script(
            """
            MailServices.accounts.removeAccount(MailServices.accounts.accounts[0], false);
            """
        )

    def subtest_simple_tab(self, opener, expected_session_data, expected_tab):
        self.marionette.set_context(self.marionette.CONTEXT_CHROME)
        self.marionette.execute_script(opener)
        self.marionette.quit(in_app=True)

        self.check_json_file([{"type": "3pane", "tabs": {"tabs": [expected_session_data]}}])

        self.marionette.start_session()
        self.marionette.set_context(self.marionette.CONTEXT_CHROME)
        self.assertEqual("mail:3pane", self.marionette.get_window_type())
        tabs = self.marionette.execute_async_script(get_tabs)
        self.assertEqual(
            [mail_3pane_tab, expected_tab],
            tabs,
        )
        self.marionette.execute_script(close_tabs)

    def test_addressbook_tab(self):
        self.subtest_simple_tab(
            'openTab("addressBookTab");',
            {"mode": "addressBookTab"},
            {"mode": "addressBookTab", "url": "about:addressbook"},
        )

    def test_calendar_tab(self):
        self.subtest_simple_tab('openTab("calendar");', {"mode": "calendar"}, {"mode": "calendar"})

    def test_tasks_tab(self):
        self.subtest_simple_tab('openTab("tasks");', {"mode": "tasks"}, {"mode": "tasks"})

    def test_chat_tab(self):
        self.subtest_simple_tab('openTab("chat");', {"mode": "chat"}, {"mode": "chat"})

    def test_content_tab(self):
        self.subtest_simple_tab(
            'openContentTab("about:mozilla");',
            {
                "mode": "contentTab",
                "state": {
                    "tabURI": "about:mozilla",
                    "linkHandler": "single-site",
                    "userContextId": "0",
                },
            },
            {
                "mode": "contentTab",
                "url": "about:mozilla",
                "linkHandler": "single-site",
                "userContextId": "0",
            },
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from marionette_driver import Wait
from marionette_harness import MarionetteTestCase

default_visible_pref = "sidebar.revamp.defaultLauncherVisible"

initial_prefs = {
    "sidebar.revamp": False,
    default_visible_pref: False,
    # Set browser restore previous session pref
    # we'll need to examine behavior using restored sidebar properties
    "browser.startup.page": 3,
}


class TestDefaultLauncherVisible(MarionetteTestCase):

    def tearDown(self):
        try:
            # Make sure subsequent tests get a clean profile
            self.marionette.restart(in_app=False, clean=True)
        finally:
            super().tearDown()

    def _close_last_tab(self):
        # "self.marionette.close" cannot be used because it doesn't
        # allow closing the very last tab.
        self.marionette.execute_script("window.close()")

    def restart_with_prefs(self, prefs):
        # set the prefs then restart the browser
        self.marionette.set_prefs(prefs)
        self.marionette.restart()

        # Restore the context as used before the restart
        self.marionette.set_context("chrome")

        self.wait_for_startup_idle_promise()

    def is_launcher_visible(self):
        hidden = self.marionette.execute_script(
            """
            const window = BrowserWindowTracker.getTopWindow();
            return window.SidebarController.sidebarContainer.hidden;
            """
        )
        return not hidden

    def is_button_visible(self):
        visible = self.marionette.execute_script(
            """
            const window = BrowserWindowTracker.getTopWindow();
            const placement = window.CustomizableUI.getPlacementOfWidget('sidebar-button');
            if (!placement) {
                return false;
            }
            const node = window.document.getElementById("sidebar-button");
            return node && !node.hidden;
            """
        )
        return visible

    def click_toolbar_button(self):
        # Click the button to show the launcher
        self.marionette.execute_script(
            """
            const window = BrowserWindowTracker.getTopWindow();
            return window.document.getElementById("sidebar-button").click()
            """
        )

    def wait_for_startup_idle_promise(self):
        self.marionette.set_context("chrome")
        self.marionette.execute_async_script(
            """
            let resolve = arguments[0];
            let { BrowserInitState } = ChromeUtils.importESModule("resource:///modules/BrowserGlue.sys.mjs");
            BrowserInitState.startupIdleTaskPromise.then(resolve);
            """
        )

    def test_first_use_default_visible_pref_false(self):
        # We flip sidebar.revamp to true, with defaultLauncherVisible=false for a profile
        # that has never enabled or seen the sidebar launcher.

        self.wait_for_startup_idle_promise()

        self.assertFalse(
            self.is_launcher_visible(),
            "Sidebar launcher is hidden",
        )
        self.assertFalse(
            self.is_button_visible(),
            "Sidebar toolbar button is hidden",
        )

        # Mimic an update which enables sidebar.revamp for the first time
        # ,with defaultLauncherVisible false
        self.restart_with_prefs(
            {
                "sidebar.revamp": True,
                "browser.startup.page": 3,
                default_visible_pref: False,
            }
        )

        Wait(self.marionette).until(
            lambda _: self.is_button_visible(),
            message="Sidebar button should be visible",
        )
        self.assertFalse(
            self.is_launcher_visible(),
            "Sidebar launcher remains hidden because defaultLauncherVisible=false",
        )
        # Click the button and verify that sticks
        self.click_toolbar_button()

        Wait(self.marionette).until(
            lambda _: self.is_launcher_visible(),
            message="Sidebar button should be visible",
        )
        self.marionette.restart()
        self.marionette.set_context("chrome")
        self.wait_for_startup_idle_promise()

        self.assertTrue(
            self.is_launcher_visible(),
            "Sidebar launcher remains visible because user un-hid it in the resumed session",
        )

    def test_new_sidebar_enabled_default_visible_pref_false(self):
        # Set up the profile with session restore and new sidebar enabled
        self.restart_with_prefs(
            {
                "sidebar.revamp": True,
                "browser.startup.page": 3,
                default_visible_pref: False,
            }
        )

        self.assertFalse(
            self.is_launcher_visible(),
            "Sidebar launcher is hidden",
        )
        self.assertTrue(
            self.is_button_visible(),
            "Sidebar toolbar button is visible",
        )

        # Mimic an update which enables sidebar.revamp with defaultLauncherVisible true
        # The restart with session restore enabled while the launcher is visible should persist
        # the launcherVisible=true and we don't want to override that
        self.restart_with_prefs(
            {
                "sidebar.revamp": True,
                "browser.startup.page": 3,
                default_visible_pref: True,
            }
        )

        Wait(self.marionette).until(
            lambda _: self.is_button_visible(),
            message="Sidebar button is still visible",
        )

        self.assertTrue(
            self.is_launcher_visible(),
            "Sidebar launcher is still visible",
        )

        # Click the button and verify that sticks
        self.click_toolbar_button()

        Wait(self.marionette).until(
            lambda _: not self.is_launcher_visible(),
            message="Sidebar launcher should be hidden",
        )
        self.marionette.restart()
        self.marionette.set_context("chrome")
        self.wait_for_startup_idle_promise()

        self.assertFalse(
            self.is_launcher_visible(),
            "Sidebar launcher remains hidden on restart",
        )

    def test_vertical_tabs_default_hidden(self):
        # Verify that starting with verticalTabs enabled and default visibility false results in a visible
        # launcher with the vertical tabstrip
        self.restart_with_prefs(
            {
                "sidebar.revamp": True,
                "sidebar.verticalTabs": True,
                default_visible_pref: False,
            }
        )

        Wait(self.marionette).until(
            lambda _: self.is_launcher_visible(),
            message="Sidebar launcher should be initially visible",
        )
        tabsWidth = self.marionette.execute_script(
            """
            const window = BrowserWindowTracker.getTopWindow();
            return document.getElementById("vertical-tabs").getBoundingClientRect().width;
            """
        )
        self.assertGreater(tabsWidth, 0, "#vertical-tabs element has width")

        # switch to 'hide-sidebar' visibility mode and confirm the launcher becomes hidden
        self.marionette.set_pref("sidebar.visibility", "hide-sidebar")
        Wait(self.marionette).until(
            lambda _: not self.is_launcher_visible(),
            message="Sidebar launcher should become hidden when hide-sidebar visibility is set and defaultLauncherVisible is false",
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import subprocess
import sys

import mozunit
from buildconfig import topsrcdir


def test_sites_compatible(tmpdir: str):
    result = subprocess.run(
        [sys.executable, "mach", "generate-python-lockfiles"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=topsrcdir,
        text=True,
    )

    # We pipe stderr to stdout and print here so that on error, the combined output
    # appears together in the test logs, making it much easier to read.
    print(result.stdout)

    assert result.returncode == 0


if __name__ == "__main__":
    mozunit.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from mach.decorators import Command


@Command("cmd_default_visible", category="testing")
def run_default_visible(self, command_context):
    pass


@Command("cmd_not_hidden", category="testing")
def run_not_hidden(self, command_context, hidden=False):
    pass


@Command("cmd_hidden", category="testing", hidden=True)
def run_hidden(self, command_context):
    pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, you can obtain one at http://mozilla.org/MPL/2.0/.

import json
import os.path

from marionette_harness import MarionetteTestCase

here = os.path.dirname(__file__)


class TestMenubarMigration(MarionetteTestCase):
    get_autohide_attribute = """
        return document.getElementById("toolbar-menubar").getAttribute("autohide");
    """

    get_xulstore_value = """
        return Services.xulStore.getValue(
            "chrome://messenger/content/messenger.xhtml",
            "toolbar-menubar",
            "autohide"
        );
    """

    get_pref_value = """
        return Services.prefs.getIntPref("mail.ui-rdf.version", -1);
    """

    def test_no_xulstore(self):
        """
        A completely new profile. The menubar should be hidden, because that's the default.
        """
        self.subtest(None, None, "true", "")

    def test_no_xulstore_no_update(self):
        """
        The XUL store has no value and migration doesn't happen. The menubar should be hidden,
        because that's the default.
        """
        self.subtest(52, None, "true", "")

    def test_no_xulstore_update(self):
        """
        The XUL store has no value when migration happens. The menubar should be visible, because
        that's what no value in the XUL store meant before ESR 140.
        """
        self.subtest(46, None, "false", "false")

    def test_xulstore_false_no_update(self):
        """
        The XUL store value is false and migration doesn't happen. The menubar should be visible.
        """
        self.subtest(52, "false", "false", "false")

    def test_xulstore_false_update(self):
        """
        The XUL store value is false when migration happens. The menubar should be visible.
        """
        self.subtest(46, "false", "false", "false")

    def test_xulstore_true_no_update(self):
        """
        The XUL store value is true and migration doesn't happen. The menubar should be hidden.
        """
        self.subtest(52, "true", "true", "true")

    def test_xulstore_true_update(self):
        """
        The XUL store value is true when migration happens. The menubar should be hidden.
        """
        self.subtest(46, "true", "true", "true")

    def subtest(
        self, migration_version, xulstore_value, expected_attribute_value, expected_xulstore_value
    ):
        self.marionette.set_context(self.marionette.CONTEXT_CHROME)
        self.marionette.quit(in_app=True)

        # Set the migration pref to a number higher than the migration we're testing
        prefs_path = os.path.join(self.marionette.profile_path, "prefs.js")
        if migration_version is None:
            os.unlink(prefs_path)
        else:
            with open(prefs_path, "w") as prefs_file:
                prefs_file.write(f"""user_pref("mail.ui-rdf.version", {migration_version});\n""")

        # Write a xulstore.json file that has the attribute set to true.
        xulstore_path = os.path.join(self.marionette.profile_path, "xulstore.json")
        if migration_version is None:
            os.unlink(xulstore_path)
        else:
            if xulstore_value is None:
                xulstore_data = {"chrome://messenger/content/messenger.xhtml": {}}
            else:
                xulstore_data = {
                    "chrome://messenger/content/messenger.xhtml": {
                        "toolbar-menubar": {"autohide": xulstore_value}
                    }
                }
            with open(xulstore_path, "w") as xulstore_file:
                json.dump(xulstore_data, xulstore_file)

        # Start the application, check the state.
        self.marionette.start_session()
        self.marionette.set_context(self.marionette.CONTEXT_CHROME)
        self.assertEqual(
            expected_attribute_value, self.marionette.execute_script(self.get_autohide_attribute)
        )
        self.assertEqual(
            expected_xulstore_value, self.marionette.execute_script(self.get_xulstore_value)
        )
        self.assertTrue(self.marionette.execute_script(self.get_pref_value) > 47)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


from marionette_driver import Wait
from marionette_harness import MarionetteTestCase


class TestLastUsedSidebarPanel(MarionetteTestCase):
    def setUp(self):
        super().setUp()
        self.marionette.set_context("chrome")
        # Add the sidebar button widget at the start of the horizontal tabstrip
        self.marionette.execute_script(
            "CustomizableUI.addWidgetToArea('sidebar-button', CustomizableUI.AREA_NAVBAR, 0)"
        )

    def tearDown(self):
        try:
            # Make sure subsequent tests get a clean profile
            self.marionette.restart(in_app=False, clean=True)
        finally:
            super().tearDown()

    def click_toolbar_button(self):
        # Click the button to show the launcher
        self.marionette.execute_script(
            """
            const window = BrowserWindowTracker.getTopWindow();
            return window.document.getElementById("sidebar-button").click()
            """
        )

    def is_sidebar_panel_visible(self):
        hidden = self.marionette.execute_script(
            """
            const window = BrowserWindowTracker.getTopWindow();
            return window.document.getElementById("sidebar-box").hidden;
            """
        )
        return not hidden

    def get_current_sidebar_id(self):
        return self.marionette.execute_script(
            """
            const window = BrowserWindowTracker.getTopWindow();
            return window.document.getElementById("sidebar-box").getAttribute("sidebarcommand");
            """
        )

    def test_panel_reopened_legacy(self):
        self.assertFalse(
            self.is_sidebar_panel_visible(), "The sidebar panel is not initially shown"
        )
        self.assertEqual(
            self.get_current_sidebar_id(), None, "The sidebar panel has no current ID"
        )

        # Open the history sidebar
        self.marionette.execute_script(
            """
            const window = BrowserWindowTracker.getTopWindow();
            return window.SidebarController.toggle("viewHistorySidebar");
            """
        )

        Wait(self.marionette).until(
            lambda _: self.get_current_sidebar_id() == "viewHistorySidebar",
            message="The history sidebar is visible",
        )

        # Click the toolbar button to close the panel
        self.click_toolbar_button()

        self.assertFalse(
            self.is_sidebar_panel_visible(), "The sidebar panel is now closed"
        )

        self.marionette.restart(clean=False, in_app=True)
        self.marionette.set_context("chrome")

        self.assertFalse(
            self.is_sidebar_panel_visible(), "The sidebar panel is initially closed"
        )
        self.click_toolbar_button()

        Wait(self.marionette).until(
            lambda _: self.get_current_sidebar_id() == "viewHistorySidebar",
            message="The history sidebar is visible",
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# add this directory to the path
import os
import sys
from urllib.parse import quote

sys.path.append(os.path.dirname(__file__))

from session_store_test_case import SessionStoreTestCase


def inline(doc):
    return f"data:text/html;charset=utf-8,{quote(doc)}"


# Each list element represents a window of tabs loaded at
# some testing URL
DEFAULT_WINDOWS = set(
    [
        # Window 1. Note the comma after the inline call -
        # this is Python's way of declaring a 1 item tuple.
        (inline("""<div">Lorem</div>"""), inline("""<div">Ipsum</div>""")),
    ]
)


class TestAutoRestoreWithTabGroups(SessionStoreTestCase):
    def setUp(self):
        super(TestAutoRestoreWithTabGroups, self).setUp(
            startup_page=3,
            include_private=False,
            restore_on_demand=True,
            test_windows=DEFAULT_WINDOWS,
        )
        self.marionette.set_context("chrome")

    def test_saved_groups_restored(self):
        self.wait_for_windows(
            self.all_windows, "Not all requested windows have been opened"
        )

        self.marionette.execute_async_script(
            """
            let resolve = arguments[0];
            let group = gBrowser.addTabGroup([gBrowser.tabs[0]], { id: "test-group", label: "test-group" });
            let { TabStateFlusher } = ChromeUtils.importESModule("resource:///modules/sessionstore/TabStateFlusher.sys.mjs");
            TabStateFlusher.flushWindow(gBrowser.ownerGlobal).then(resolve);
            """
        )

        self.assertEqual(
            self.marionette.execute_script("return gBrowser.getAllTabGroups().length"),
            1,
            "There is one open group",
        )

        self.marionette.quit()
        self.marionette.start_session()
        self.marionette.set_context("chrome")

        self.assertEqual(
            self.marionette.execute_script("return gBrowser.getAllTabGroups().length"),
            1,
            "There is one open group",
        )

        self.assertEqual(
            self.marionette.execute_script("return SessionStore.savedGroups.length"),
            0,
            "The group was not saved because it was automatically restored",
        )

        self.marionette.execute_script(
            """
            let group = gBrowser.getTabGroupById("test-group");
            group.ownerGlobal.SessionStore.addSavedTabGroup(group);
            group.ownerGlobal.gBrowser.removeTabGroup(group, { animate: false });
            """
        )

        self.assertEqual(
            self.marionette.execute_script("return SessionStore.savedGroups.length"),
            1,
            "The group is now saved",
        )

        self.marionette.quit()
        self.marionette.start_session()
        self.marionette.set_context("chrome")

        self.assertEqual(
            self.marionette.execute_script("return gBrowser.getAllTabGroups().length"),
            0,
            "The group was not automatically restored because it was manually saved",
        )

        self.assertEqual(
            self.marionette.execute_script("return SessionStore.savedGroups.length"),
            1,
            "The saved group persists after a second restart",
        )

        self.marionette.execute_script(
            """
            SessionStore.forgetSavedTabGroup("test-group");
            """
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 0.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/0.0/.

from marionette_harness import MarionetteTestCase


class TestNoToolbarChanges(MarionetteTestCase):
    """
    Test that toolbar widgets remain in the same order over several restarts of the browser
    """

    have_seen_import_button = False

    def setUp(self):
        super().setUp()
        self.marionette.set_context("chrome")

    def get_area_widgets(self, area):
        return self.marionette.execute_script(
            f"return CustomizableUI.getWidgetIdsInArea(CustomizableUI.{area}).map(id => id.includes('spring') ? 'spring' : id)"
        )

    def get_area_default_placements(self, area):
        return self.marionette.execute_script(
            f"return CustomizableUI.getDefaultPlacementsForArea(CustomizableUI.{area})"
        )

    def check_toolbar_placements(self):
        self.assertEqual(
            self.get_area_widgets("AREA_TABSTRIP"),
            self.get_area_default_placements("AREA_TABSTRIP"),
            msg="AREA_TABSTRIP placements are as expected",
        )
        navbarPlacements = self.get_area_default_placements("AREA_NAVBAR")
        navbarPlacements.append("unified-extensions-button")
        self.assertEqual(
            self.get_area_widgets("AREA_NAVBAR"),
            navbarPlacements,
            msg="AREA_NAVBAR placements are as expected",
        )
        actualBookmarkPlacements = self.get_area_widgets("AREA_BOOKMARKS")
        bookmarkPlacements = self.get_area_default_placements("AREA_BOOKMARKS")
        # The import button is added lazily on startup, so we can't predict
        # whether it'll be here. Turning it off via prefs=[] annotations on the
        # test also doesn't work
        # (https://bugzilla.mozilla.org/show_bug.cgi?id=1959688).
        # So we simply accept placements either with or without the button - but
        # if we ever see the button we should keep seeing it.
        self.have_seen_import_button = (
            self.have_seen_import_button or "import-button" in actualBookmarkPlacements
        )
        if self.have_seen_import_button:
            self.assertEqual(
                actualBookmarkPlacements,
                ["import-button"] + bookmarkPlacements,
                msg="AREA_BOOKMARKS placements are as expected",
            )
        else:
            self.assertEqual(
                actualBookmarkPlacements,
                bookmarkPlacements,
                msg="AREA_BOOKMARKS placements are as expected",
            )

        self.assertEqual(
            self.get_area_widgets("AREA_ADDONS"),
            self.get_area_default_placements("AREA_ADDONS"),
            msg="AREA_ADDONS placements are as expected",
        )
        self.assertEqual(
            self.get_area_widgets("AREA_FIXED_OVERFLOW_PANEL"),
            self.get_area_default_placements("AREA_FIXED_OVERFLOW_PANEL"),
            msg="AREA_FIXED_OVERFLOW_PANEL placements are as expected",
        )

    def test_no_toolbar_changes(self):
        self.check_toolbar_placements()
        self.marionette.restart()
        self.check_toolbar_placements()
        self.marionette.restart()
        self.check_toolbar_placements()
        self.marionette.restart()
        self.check_toolbar_placements()
        self.marionette.restart()
        self.check_toolbar_placements()
        self.marionette.restart()
        self.check_toolbar_placements()



#!/usr/bin/env python
# coding=UTF-8

from os import walk

import mozunit
import pytest
from conftest import fspath


def test_no_dump_files(check_for_crashes):
    """Test that check_for_crashes returns 0 if no dumps are present."""
    assert 0 == check_for_crashes()


@pytest.mark.parametrize("minidump_files", [3], indirect=True)
def test_dump_count(check_for_crashes, minidump_files):
    """Test that check_for_crashes returns the number of crash dumps."""
    assert 3 == check_for_crashes()


def test_dump_directory_unicode(request, check_for_crashes, tmpdir, capsys):
    """Test that check_for_crashes can handle unicode in dump_directory."""
    from conftest import minidump_files

    tmpdir = tmpdir.ensure("🍪", dir=1)
    minidump_files = minidump_files(request, tmpdir)

    assert 1 == check_for_crashes(dump_directory=fspath(tmpdir), quiet=False)

    out, _ = capsys.readouterr()
    assert fspath(minidump_files[0]["dmp"]) in out
    assert "🍪" in out


def test_test_name_unicode(check_for_crashes, minidump_files, capsys):
    """Test that check_for_crashes can handle unicode in dump_directory."""
    assert 1 == check_for_crashes(test_name="🍪", quiet=False)

    out, err = capsys.readouterr()
    assert "| 🍪" in out


@pytest.mark.parametrize("keep", [True, False, None])
def test_minidump_files_are_cleaned_up_or_preserved_in_original_location(
    request, check_for_crashes, tmpdir, keep
):
    from conftest import minidump_files

    tmpdir = tmpdir.ensure("test", dir=1)
    minidump_files = minidump_files(request, tmpdir)

    # Make sure that minidump files are present in the temporary location.
    minidump_files_in_dir = next(walk(tmpdir), (None, None, []))[2]
    assert len(minidump_files_in_dir) == 2

    check_for_crashes(dump_directory=fspath(tmpdir), quiet=False, keep=keep)

    # Make sure that minidump files are preserved if keep=True or
    # removed otherwise.
    minidump_files_in_dir = next(walk(tmpdir), (None, None, []))[2]
    assert len(minidump_files_in_dir) == (2 if keep is True else 0)


if __name__ == "__main__":
    mozunit.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from telemetry_harness.fog_testcase import FOGTestCase


class BaselineRidealongFilter:
    """Expecting a `baseline` ping and one more ping"""

    def __init__(self, other_ping):
        self.expected_pings = ["baseline", other_ping]

    def __call__(self, ping):
        doc_type = ping["request_url"]["doc_type"]
        return doc_type in self.expected_pings


DauReportFilter = BaselineRidealongFilter("usage-reporting")


class TestClientActivity(FOGTestCase):
    """
    Tests for client activity and FOG's scheduling of the "baseline" ping.
    For every `baseline` ping we also expect a `usage-reporting` ping.
    """

    def test_user_activity(self):
        # We might still get those pings in an unexpected order.
        # Let's check we get both.
        expected_pings = ["baseline", "usage-reporting"]

        # Restarting the browser could send multiple "baseline" pings
        # (e.g. there could be some pending, or some triggered near shutdown.)
        # So only accept the "baseline" ping sent near startup.
        # We identify this ping as the one that has seen an "events" ping being sent,
        # since one of those is submitted during startup due to events being present.
        def is_startup_baseline_ping(ping):
            return (
                ping["request_url"]["doc_type"] == "baseline"
                and ping["payload"]["metrics"]
                .get("labeled_counter", {})
                .get("glean.validation.pings_submitted", {})
                .get("events")
                == 1
            )

        [ping0, ping1] = self.wait_for_pings(
            self.restart_browser,
            lambda ping: is_startup_baseline_ping(ping)
            or ping["request_url"]["doc_type"] == "usage-reporting",
            2,
            ping_server=self.fog_ping_server,
        )

        # First test that restarting the browser sends a "active" ping
        received_pings = sorted(
            [
                ping0["request_url"]["doc_type"],
                ping1["request_url"]["doc_type"],
            ]
        )
        self.assertEqual(expected_pings, received_pings)
        if ping0["request_url"]["doc_type"] == "baseline":
            self.assertEqual("active", ping0["payload"]["ping_info"]["reason"])
        if ping1["request_url"]["doc_type"] == "baseline":
            self.assertEqual("active", ping1["payload"]["ping_info"]["reason"])

        with self.marionette.using_context(self.marionette.CONTEXT_CHROME):
            zero_prefs_script = """\
            Services.prefs.setIntPref("telemetry.fog.test.inactivity_limit", 0);
            Services.prefs.setIntPref("telemetry.fog.test.activity_limit", 0);
            """
            self.marionette.execute_script(zero_prefs_script)

        def user_active(active, marionette):
            script = "Services.obs.notifyObservers(null, 'user-interaction-{}active')".format(
                "" if active else "in"
            )
            with marionette.using_context(marionette.CONTEXT_CHROME):
                marionette.execute_script(script)

        [ping2, ping3] = self.wait_for_pings(
            lambda: user_active(True, self.marionette),
            DauReportFilter,
            2,
            ping_server=self.fog_ping_server,
        )

        [ping4, ping5] = self.wait_for_pings(
            lambda: user_active(False, self.marionette),
            DauReportFilter,
            2,
            ping_server=self.fog_ping_server,
        )

        received_pings = sorted(
            [
                ping2["request_url"]["doc_type"],
                ping3["request_url"]["doc_type"],
            ]
        )
        self.assertEqual(expected_pings, received_pings)
        if ping2["request_url"]["doc_type"] == "baseline":
            self.assertEqual("active", ping2["payload"]["ping_info"]["reason"])
        if ping3["request_url"]["doc_type"] == "baseline":
            self.assertEqual("active", ping3["payload"]["ping_info"]["reason"])

        received_pings = sorted(
            [
                ping4["request_url"]["doc_type"],
                ping5["request_url"]["doc_type"],
            ]
        )
        self.assertEqual(expected_pings, received_pings)
        if ping4["request_url"]["doc_type"] == "baseline":
            self.assertEqual("inactive", ping4["payload"]["ping_info"]["reason"])
        if ping5["request_url"]["doc_type"] == "baseline":
            self.assertEqual("inactive", ping5["payload"]["ping_info"]["reason"])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import ast
import functools
import importlib.metadata
import os
import subprocess
import sys
import tempfile
from pathlib import Path
from subprocess import CompletedProcess
from typing import List

import buildconfig
import mozunit
import pytest

from mach.site import MozSiteMetadata, PythonVirtualenv, activate_virtualenv


class ActivationContext:
    def __init__(
        self,
        topsrcdir: Path,
        work_dir: Path,
        original_python_path: str,
        stdlib_paths: List[Path],
        system_paths: List[Path],
        required_mach_sys_paths: List[Path],
        mach_requirement_paths: List[Path],
        command_requirement_path: Path,
    ):
        self.topsrcdir = topsrcdir
        self.work_dir = work_dir
        self.original_python_path = original_python_path
        self.stdlib_paths = stdlib_paths
        self.system_paths = system_paths
        self.required_moz_init_sys_paths = required_mach_sys_paths
        self.mach_requirement_paths = mach_requirement_paths
        self.command_requirement_path = command_requirement_path

    def virtualenv(self, name: str) -> PythonVirtualenv:
        base_path = self.work_dir

        if name == "mach":
            base_path = base_path / "_virtualenvs"
        return PythonVirtualenv(str(base_path / name))


def test_new_package_metadta_is_found():
    try:
        # "carrot" was chosen as the package to use because:
        # * It has to be a package that doesn't exist in-scope at the start (so,
        #   all vendored modules included in the test virtualenv aren't usage).
        # * It must be on our internal PyPI mirror.
        # Of the options, "carrot" is a small install that fits these requirements.
        importlib.metadata.distribution("carrot")
        assert False, "Expected to not find 'carrot' as the initial state of the test"
    except importlib.metadata.PackageNotFoundError:
        pass

    with tempfile.TemporaryDirectory() as venv_dir:
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "venv",
                venv_dir,
            ]
        )

        venv = PythonVirtualenv(venv_dir)
        venv.pip_install(["carrot==0.10.7"])

        initial_metadata = MozSiteMetadata.from_runtime()
        try:
            metadata = MozSiteMetadata(None, None, None, None, venv.prefix)
            with metadata.update_current_site(venv.python_path):
                activate_virtualenv(venv)

            assert importlib.metadata.distribution("carrot").version == "0.10.7"
        finally:
            MozSiteMetadata.current = initial_metadata


def test_sys_path_source_none_build(context):
    original, mach, command = _run_activation_script_for_paths(context, "none", "build")
    _assert_original_python_sys_path(context, original)

    assert not os.path.exists(context.virtualenv("mach").prefix)
    assert mach == [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
    ]

    expected_command_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        context.command_requirement_path,
    ]
    assert command == expected_command_paths

    command_venv = _sys_path_of_virtualenv(context.virtualenv("build"))
    assert command_venv == [Path(""), *expected_command_paths]


def test_sys_path_source_none_other(context):
    original, mach, command = _run_activation_script_for_paths(context, "none", "other")
    _assert_original_python_sys_path(context, original)

    assert not os.path.exists(context.virtualenv("mach").prefix)
    assert mach == [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
    ]

    command_virtualenv = PythonVirtualenv(str(context.work_dir / "other"))
    expected_command_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        context.command_requirement_path,
        *(Path(p) for p in command_virtualenv.site_packages_dirs()),
    ]
    assert command == expected_command_paths

    command_venv = _sys_path_of_virtualenv(context.virtualenv("other"))
    assert command_venv == [
        Path(""),
        *expected_command_paths,
    ]


def test_sys_path_source_venv_build(context):
    original, mach, command = _run_activation_script_for_paths(context, "pip", "build")
    _assert_original_python_sys_path(context, original)

    mach_virtualenv = context.virtualenv("mach")
    expected_mach_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        *(Path(p) for p in mach_virtualenv.site_packages_dirs()),
    ]
    assert mach == expected_mach_paths

    command_virtualenv = context.virtualenv("build")
    expected_command_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        *(Path(p) for p in mach_virtualenv.site_packages_dirs()),
        context.command_requirement_path,
        *(Path(p) for p in command_virtualenv.site_packages_dirs()),
    ]
    assert command == expected_command_paths

    mach_venv = _sys_path_of_virtualenv(mach_virtualenv)
    assert mach_venv == [
        Path(""),
        *expected_mach_paths,
    ]

    command_venv = _sys_path_of_virtualenv(command_virtualenv)
    assert command_venv == [
        Path(""),
        *expected_command_paths,
    ]


def test_sys_path_source_venv_other(context):
    original, mach, command = _run_activation_script_for_paths(context, "pip", "other")
    _assert_original_python_sys_path(context, original)

    mach_virtualenv = context.virtualenv("mach")
    expected_mach_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        *(Path(p) for p in mach_virtualenv.site_packages_dirs()),
    ]
    assert mach == expected_mach_paths

    command_virtualenv = context.virtualenv("other")
    expected_command_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        *(Path(p) for p in mach_virtualenv.site_packages_dirs()),
        context.command_requirement_path,
        *(Path(p) for p in command_virtualenv.site_packages_dirs()),
    ]
    assert command == expected_command_paths

    mach_venv = _sys_path_of_virtualenv(mach_virtualenv)
    assert mach_venv == [
        Path(""),
        *expected_mach_paths,
    ]

    command_venv = _sys_path_of_virtualenv(command_virtualenv)
    assert command_venv == [
        Path(""),
        *expected_command_paths,
    ]


def test_sys_path_source_system_build(context):
    original, mach, command = _run_activation_script_for_paths(
        context, "system", "build"
    )
    _assert_original_python_sys_path(context, original)

    assert not os.path.exists(context.virtualenv("mach").prefix)
    expected_mach_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        *context.system_paths,
    ]
    assert mach == expected_mach_paths

    command_virtualenv = context.virtualenv("build")
    expected_command_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        *context.system_paths,
        context.command_requirement_path,
    ]
    assert command == expected_command_paths

    command_venv = _sys_path_of_virtualenv(command_virtualenv)
    assert command_venv == [
        Path(""),
        *expected_command_paths,
    ]


def test_sys_path_source_system_other(context):
    result = _run_activation_script(
        context,
        "system",
        "other",
        context.original_python_path,
        stderr=subprocess.PIPE,
    )
    assert result.returncode != 0
    assert (
        'Cannot use MACH_BUILD_PYTHON_NATIVE_PACKAGE_SOURCE="system" for any sites '
        "other than" in result.stderr
    )


def test_sys_path_source_venvsystem_build(context):
    venv_system_python = _create_venv_system_python(
        context.work_dir, context.original_python_path
    )
    venv_system_site_packages_dirs = [
        Path(p) for p in venv_system_python.site_packages_dirs()
    ]
    original, mach, command = _run_activation_script_for_paths(
        context, "system", "build", venv_system_python.python_path
    )

    assert original == [
        Path(__file__).parent,
        *context.required_moz_init_sys_paths,
        *context.stdlib_paths,
        *venv_system_site_packages_dirs,
    ]

    assert not os.path.exists(context.virtualenv("mach").prefix)
    expected_mach_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        *venv_system_site_packages_dirs,
    ]
    assert mach == expected_mach_paths

    command_virtualenv = context.virtualenv("build")
    expected_command_paths = [
        *context.stdlib_paths,
        *context.mach_requirement_paths,
        *venv_system_site_packages_dirs,
        context.command_requirement_path,
    ]
    assert command == expected_command_paths

    command_venv = _sys_path_of_virtualenv(command_virtualenv)
    assert command_venv == [
        Path(""),
        *expected_command_paths,
    ]


def test_sys_path_source_venvsystem_other(context):
    venv_system_python = _create_venv_system_python(
        context.work_dir, context.original_python_path
    )
    result = _run_activation_script(
        context,
        "system",
        "other",
        venv_system_python.python_path,
        stderr=subprocess.PIPE,
    )
    assert result.returncode != 0
    assert (
        'Cannot use MACH_BUILD_PYTHON_NATIVE_PACKAGE_SOURCE="system" for any sites '
        "other than" in result.stderr
    )


@pytest.fixture(name="context")
def _activation_context():
    original_python_path, stdlib_paths, system_paths = _original_python()
    topsrcdir = Path(buildconfig.topsrcdir)
    required_mach_sys_paths = [
        topsrcdir / "python" / "mach",
        topsrcdir / "third_party" / "python" / "filelock",
        topsrcdir / "third_party" / "python" / "packaging",
        topsrcdir / "third_party" / "python" / "pip",
    ]

    with tempfile.TemporaryDirectory() as work_dir:
        # Get "resolved" version of path to ease comparison against "site"-added sys.path
        # entries, as "site" calculates the realpath of provided locations.
        work_dir = Path(work_dir).resolve()
        mach_requirement_paths = [
            *required_mach_sys_paths,
            work_dir / "mach_site_path",
        ]
        command_requirement_path = work_dir / "command_site_path"
        (work_dir / "mach_site_path").touch()
        command_requirement_path.touch()
        yield ActivationContext(
            topsrcdir,
            work_dir,
            original_python_path,
            stdlib_paths,
            system_paths,
            required_mach_sys_paths,
            mach_requirement_paths,
            command_requirement_path,
        )


@functools.lru_cache(maxsize=None)
def _original_python():
    current_site = MozSiteMetadata.from_runtime()
    stdlib_paths, system_paths = current_site.original_python.sys_path()
    stdlib_paths = [Path(path) for path in _filter_pydev_from_paths(stdlib_paths)]
    system_paths = [Path(path) for path in system_paths]
    return current_site.original_python.python_path, stdlib_paths, system_paths


def _run_activation_script(
    context: ActivationContext,
    source: str,
    site_name: str,
    invoking_python: str,
    **kwargs
) -> CompletedProcess:
    return subprocess.run(
        [
            invoking_python,
            str(Path(__file__).parent / "script_site_activation.py"),
        ],
        stdout=subprocess.PIPE,
        text=True,
        env={
            "TOPSRCDIR": str(context.topsrcdir),
            "COMMAND_SITE": site_name,
            "PYTHONPATH": os.pathsep.join(
                str(p) for p in context.required_moz_init_sys_paths
            ),
            "MACH_SITE_PTH_REQUIREMENTS": os.pathsep.join(
                str(p) for p in context.mach_requirement_paths
            ),
            "COMMAND_SITE_PTH_REQUIREMENTS": str(context.command_requirement_path),
            "MACH_BUILD_PYTHON_NATIVE_PACKAGE_SOURCE": source,
            "WORK_DIR": str(context.work_dir),
            # These two variables are needed on Windows so that Python initializes
            # properly and adds the "user site packages" to the sys.path like normal.
            "SYSTEMROOT": os.environ.get("SYSTEMROOT", ""),
            "APPDATA": os.environ.get("APPDATA", ""),
        },
        **kwargs,
    )


def _run_activation_script_for_paths(
    context: ActivationContext, source: str, site_name: str, invoking_python: str = None
) -> List[List[Path]]:
    """Return the states of the sys.path when activating Mach-managed sites

    Three sys.path states are returned:
    * The initial sys.path, equivalent to "path_to_python -c "import sys; print(sys.path)"
    * The sys.path after activating the Mach site
    * The sys.path after activating the command site
    """

    output = _run_activation_script(
        context,
        source,
        site_name,
        invoking_python or context.original_python_path,
        check=True,
    ).stdout
    # Filter to the last line, which will have our nested list that we want to
    # parse. This will avoid unrelated output, such as from virtualenv creation
    output = output.splitlines()[-1]
    return [
        [Path(path) for path in _filter_pydev_from_paths(paths)]
        for paths in ast.literal_eval(output)
    ]


def _assert_original_python_sys_path(context: ActivationContext, original: List[Path]):
    # Assert that initial sys.path (prior to any activations) matches expectations.
    assert original == [
        Path(__file__).parent,
        *context.required_moz_init_sys_paths,
        *context.stdlib_paths,
        *context.system_paths,
    ]


def _sys_path_of_virtualenv(virtualenv: PythonVirtualenv) -> List[Path]:
    output = subprocess.run(
        [virtualenv.python_path, "-c", "import sys; print(sys.path)"],
        stdout=subprocess.PIPE,
        text=True,
        env={
            # Needed for python to initialize properly
            "SYSTEMROOT": os.environ.get("SYSTEMROOT", ""),
        },
        check=True,
    ).stdout
    return [Path(path) for path in _filter_pydev_from_paths(ast.literal_eval(output))]


def _filter_pydev_from_paths(paths: List[str]) -> List[str]:
    # Filter out injected "pydev" debugging tool if running within a JetBrains
    # debugging context.
    return [path for path in paths if "pydev" not in path and "JetBrains" not in path]


def _create_venv_system_python(
    work_dir: Path, invoking_python: str
) -> PythonVirtualenv:
    virtualenv = PythonVirtualenv(str(work_dir / "system_python"))
    subprocess.run(
        [
            invoking_python,
            "-m",
            "venv",
            virtualenv.prefix,
            "--without-pip",
        ],
        check=True,
    )
    return virtualenv


if __name__ == "__main__":
    mozunit.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


import os
import sys

from marionette_driver import Wait

# add this directory to the path
sys.path.append(os.path.dirname(__file__))
from session_store_test_case import SessionStoreTestCase


def inline(title):
    return f"data:text/html;charset=utf-8,<html><head><title>{title}</title></head><body></body></html>"


class TestManualRestoreWithTaskbarTabs(SessionStoreTestCase):
    def setUp(self):
        super(TestManualRestoreWithTaskbarTabs, self).setUp(
            startup_page=1,
            include_private=False,
            restore_on_demand=False,
            taskbartabs_enable=True,
            test_windows=set(
                [
                    # Window 1
                    (
                        inline("lorem ipsom"),
                        inline("dolor"),
                    ),
                ]
            ),
        )

    """
    Close all regular windows except for a taskbar tab window. The
    session should be over at this point. Opening another regular Firefox
    window will open "restore previous session" in the hamburger menu.
    And clicking it will restore the correct session.
    """

    def test_restore_without_closing_taskbartab(self):
        self.wait_for_windows(
            self.all_windows, "Not all requested windows have been opened"
        )

        # See session_store_test_case.py
        self.setup_taskbartab_restore_scenario()

        # Verify that "restore previous session" button
        # is visible in the hamburger menu
        self.assertEqual(
            self.marionette.execute_script(
                """
                let newWindow = BrowserWindowTracker.getTopWindow({ allowTaskbarTabs: false });
                return PanelMultiView.getViewNode(
                    newWindow.document,
                    "appMenu-restoreSession"
                ).hasAttribute("disabled");
            """
            ),
            False,
            "The restore last session button should be visible",
        )

        # Simulate clicking "restore previous session"
        self.marionette.execute_script(
            """
                SessionStore.restoreLastSession();
            """
        )

        # Wait for the restore to be completed,
        # meaning the window we opened should have
        # two tabs again.
        Wait(self.marionette).until(
            lambda mn: mn.execute_script(
                """
                let newWindow = BrowserWindowTracker.getTopWindow({ allowTaskbarTabs: false });
                return newWindow.gBrowser.tabs.length;
                """
            )
            == 2
        )


class TestAutoRestoreWithTaskbarTabs(SessionStoreTestCase):
    def setUp(self):
        super(TestAutoRestoreWithTaskbarTabs, self).setUp(
            startup_page=3,
            include_private=False,
            restore_on_demand=False,
            taskbartabs_enable=True,
            test_windows=set(
                [
                    # Window 1
                    (
                        inline("lorem ipsom"),
                        inline("dolor"),
                    ),
                ]
            ),
        )

    """
    Close all regular windows except for a taskbar tab window. The
    session should be over at this point. Opening another regular Firefox
    window will open automatically restore the correct session
    """

    def test_restore_without_closing_taskbartab(self):
        self.wait_for_windows(
            self.all_windows, "Not all requested windows have been opened"
        )

        self.setup_taskbartab_restore_scenario()

        # Wait for the auto restore to be completed,
        # meaning the window we opened should have
        # the original two tabs plus the home page tab.
        Wait(self.marionette).until(
            lambda mn: mn.execute_script(
                """
                let newWindow = BrowserWindowTracker.getTopWindow({ allowTaskbarTabs: false });
                return newWindow.gBrowser.tabs.length;
                """
            )
            == 3
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import sys
from pathlib import Path
from typing import Union
from unittest.mock import Mock

from mach.site import MozSiteMetadata, SitePackagesSource


class NoopTelemetry:
    def __init__(self, failed_glean_import):
        self._failed_glean_import = failed_glean_import

    def metrics(self, metrics_path: Union[str, Path]):
        return Mock()

    def submit(self, is_bootstrap):
        if self._failed_glean_import and not is_bootstrap:
            active_site = MozSiteMetadata.from_runtime()
            if active_site.mach_site_packages_source == SitePackagesSource.SYSTEM:
                hint = (
                    "Mach is looking for glean in the system packages. This can be "
                    "resolved by installing it there, or by allowing Mach to run "
                    "without using the system Python packages."
                )
            elif active_site.mach_site_packages_source == SitePackagesSource.NONE:
                hint = (
                    "This is because Mach is currently configured without a source "
                    "for native Python packages."
                )
            else:
                hint = "You may need to run |mach bootstrap|."

            print(
                f"Glean could not be found, so telemetry will not be reported. {hint}",
                file=sys.stderr,
            )


class GleanTelemetry:
    """Records and sends Telemetry using Glean.

    Metrics are defined in python/mozbuild/metrics.yaml.
    Pings are defined in python/mozbuild/pings.yaml.

    The "metrics" and "pings" properties may be replaced with no-op implementations if
    Glean isn't available. This allows consumers to report telemetry without having
    to guard against incompatible environments.
    """

    def __init__(
        self,
        upload_enabled: bool,
        data_dir: Path,
    ):
        from glean import Glean, load_pings

        self._metrics_cache = {}
        # We must call load_pings before Glean.initialize to load and register pings;
        # otherwise, metrics won’t be recorded for them.
        self._pings = load_pings(Path(__file__).parent.parent / "pings.yaml")

        Glean.initialize(
            application_id="mozilla.mach",
            application_version="Unknown",
            upload_enabled=upload_enabled,
            data_dir=data_dir,
        )

    def metrics(self, metrics_path: Union[str, Path]):
        if metrics_path not in self._metrics_cache:
            from glean import load_metrics

            metrics = load_metrics(metrics_path)
            self._metrics_cache[metrics_path] = metrics

        return self._metrics_cache[metrics_path]

    def submit(self, _):
        self._pings.usage.submit()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from telemetry_harness.fog_ping_filters import (
    FOG_DELETION_REQUEST_PING,
    FOGDocTypePingFilter,
)
from telemetry_harness.fog_testcase import FOGTestCase

BASELINE = FOGDocTypePingFilter("baseline")
FOG_USAGE_REPORTING = FOGDocTypePingFilter("usage-reporting")
FOG_USAGE_DELETION_REQUEST_PING = FOGDocTypePingFilter("usage-deletion-request")
CANARY_USAGE_PROFILE_ID = "beefbeef-beef-beef-beef-beeefbeefbee"
CANARY_USAGE_PROFILE_GROUP_ID = "b0bacafe-b0ba-cafe-b0ba-cafeb0bacafe"


class TestUsageReporting(FOGTestCase):
    """Tests for FOG usage-reporting ping and usage-id cycling."""

    def disable_usage_reporting(self):
        """Disable usage reporting in the current browser."""
        self.marionette.instance.profile.set_persistent_preferences(
            {"datareporting.usage.uploadEnabled": False}
        )
        self.marionette.set_pref("datareporting.usage.uploadEnabled", False)

    def enable_usage_reporting(self):
        """Enable usage reporting in the current browser."""
        self.marionette.instance.profile.set_persistent_preferences(
            {"datareporting.usage.uploadEnabled": True}
        )
        self.marionette.set_pref("datareporting.usage.uploadEnabled", True)

    def test_deletion_request(self):
        """
        Test the "usage-reporting" ping behaviour and usage-id cycling when disabling telemetry in general.

        We do not expect a "usage-deletion-request" ping.  The "deletion-request" ping should not include the usage-id.
        """

        ping1 = self.wait_for_ping(
            self.restart_browser,
            FOG_USAGE_REPORTING,
            ping_server=self.fog_ping_server,
        )

        self.assertNotIn("ping_info", ping1["payload"])
        self.assertNotIn("client_info", ping1["payload"])

        metrics = ping1["payload"]["metrics"]
        self.assertNotIn("legacy.telemetry.client_id", metrics["uuid"])
        self.assertNotIn("legacy.telemetry.profile_group_id", metrics["uuid"])

        self.assertIn("usage.profile_id", metrics["uuid"])
        usage_id1 = metrics["uuid"]["usage.profile_id"]
        self.assertIsValidUUID(usage_id1)
        self.assertNotEqual(CANARY_USAGE_PROFILE_ID, usage_id1)

        self.assertIn("usage.profile_group_id", metrics["uuid"])
        usage_group_id1 = metrics["uuid"]["usage.profile_group_id"]
        self.assertIsValidUUID(usage_group_id1)
        self.assertNotEqual(CANARY_USAGE_PROFILE_GROUP_ID, usage_group_id1)

        # Regular `deletion-request` ping won't have the `usage.profile_id`.
        # We just wait for it to know it happened.
        _ping2 = self.wait_for_ping(
            self.disable_telemetry,
            FOG_DELETION_REQUEST_PING,
            ping_server=self.fog_ping_server,
        )

        self.enable_telemetry()
        ping3 = self.wait_for_ping(
            self.restart_browser,
            FOG_USAGE_REPORTING,
            ping_server=self.fog_ping_server,
        )

        self.assertIn("usage.profile_id", ping3["payload"]["metrics"]["uuid"])
        usage_id2 = ping3["payload"]["metrics"]["uuid"]["usage.profile_id"]
        self.assertIsValidUUID(usage_id2)

        self.assertNotEqual(usage_id1, usage_id2)
        self.assertNotEqual(CANARY_USAGE_PROFILE_ID, usage_id2)

        self.assertIn("usage.profile_group_id", ping3["payload"]["metrics"]["uuid"])
        usage_group_id2 = ping3["payload"]["metrics"]["uuid"]["usage.profile_group_id"]
        self.assertIsValidUUID(usage_group_id2)

        self.assertNotEqual(usage_group_id1, usage_group_id2)
        self.assertNotEqual(CANARY_USAGE_PROFILE_GROUP_ID, usage_group_id2)

    def test_usage_deletion_request(self):
        """
        Test the "usage-reporting" ping behaviour and usage-id cycling when disabling telemetry.

        We expect a "usage-deletion-request" ping, and it should include the usage-id.
        """

        ping1 = self.wait_for_ping(
            self.restart_browser,
            FOG_USAGE_REPORTING,
            ping_server=self.fog_ping_server,
        )

        self.assertNotIn("ping_info", ping1["payload"])
        self.assertNotIn("client_info", ping1["payload"])

        metrics = ping1["payload"]["metrics"]
        self.assertNotIn("legacy.telemetry.client_id", metrics["uuid"])
        self.assertNotIn("legacy.telemetry.profile_group_id", metrics["uuid"])

        self.assertIn("usage.profile_id", metrics["uuid"])
        usage_id1 = metrics["uuid"]["usage.profile_id"]
        self.assertIsValidUUID(usage_id1)
        self.assertNotEqual(CANARY_USAGE_PROFILE_ID, usage_id1)

        self.assertIn("usage.profile_group_id", metrics["uuid"])
        usage_group_id1 = metrics["uuid"]["usage.profile_group_id"]
        self.assertIsValidUUID(usage_group_id1)
        self.assertNotEqual(CANARY_USAGE_PROFILE_GROUP_ID, usage_group_id1)

        # `usage-deletion-request` ping will have the `usage.profile_id`.
        #
        # N.b.: the `usage-deletion-request` ping has `include_info_sections:
        # false`, so no `payload.ping_info.reason` is available for inspection.
        # We still set the reason in case details change.
        ping2 = self.wait_for_ping(
            self.disable_usage_reporting,
            FOG_USAGE_DELETION_REQUEST_PING,
            ping_server=self.fog_ping_server,
        )
        metrics = ping2["payload"]["metrics"]
        self.assertIn("usage.profile_id", metrics["uuid"])
        usage_id2 = metrics["uuid"]["usage.profile_id"]
        self.assertEqual(usage_id1, usage_id2)

        self.assertIn("usage.profile_group_id", metrics["uuid"])
        usage_group_id2 = metrics["uuid"]["usage.profile_group_id"]
        self.assertEqual(usage_group_id1, usage_group_id2)

        self.enable_usage_reporting()
        ping3 = self.wait_for_ping(
            self.restart_browser,
            FOG_USAGE_REPORTING,
            ping_server=self.fog_ping_server,
        )

        self.assertIn("usage.profile_id", ping3["payload"]["metrics"]["uuid"])
        usage_id3 = ping3["payload"]["metrics"]["uuid"]["usage.profile_id"]
        self.assertIsValidUUID(usage_id3)

        self.assertNotEqual(usage_id1, usage_id3)
        self.assertNotEqual(CANARY_USAGE_PROFILE_ID, usage_id3)

        self.assertIn("usage.profile_group_id", ping3["payload"]["metrics"]["uuid"])
        usage_group_id3 = ping3["payload"]["metrics"]["uuid"]["usage.profile_group_id"]
        self.assertIsValidUUID(usage_group_id3)

        self.assertNotEqual(usage_group_id1, usage_group_id3)
        self.assertNotEqual(CANARY_USAGE_PROFILE_GROUP_ID, usage_group_id3)

    def test_enabled_state_after_restart(self):
        """
        Test that the "usage-reporting" ping remains enabled and the usage ID and usage group ID remain fixed when restarting the browser.
        """

        self.disable_usage_reporting()
        # Not guaranteed to send a "usage-reporting" ping.
        self.enable_usage_reporting()

        # But restarting should send a "usage-reporting ping".
        ping1 = self.wait_for_ping(
            self.restart_browser,
            FOG_USAGE_REPORTING,
            ping_server=self.fog_ping_server,
        )

        metrics = ping1["payload"]["metrics"]
        self.assertIn("usage.profile_id", metrics["uuid"])
        usage_id1 = metrics["uuid"]["usage.profile_id"]
        self.assertIsValidUUID(usage_id1)
        self.assertNotEqual(CANARY_USAGE_PROFILE_ID, usage_id1)

        self.assertIn("usage.profile_group_id", metrics["uuid"])
        usage_group_id1 = metrics["uuid"]["usage.profile_group_id"]
        self.assertIsValidUUID(usage_group_id1)
        self.assertNotEqual(CANARY_USAGE_PROFILE_GROUP_ID, usage_group_id1)

        # Restarting again should maintain enabled state.
        ping2 = self.wait_for_ping(
            self.restart_browser,
            FOG_USAGE_REPORTING,
            ping_server=self.fog_ping_server,
        )

        metrics = ping2["payload"]["metrics"]
        self.assertIn("usage.profile_id", metrics["uuid"])
        usage_id2 = metrics["uuid"]["usage.profile_id"]

        self.assertEqual(usage_id1, usage_id2)

        self.assertIn("usage.profile_group_id", metrics["uuid"])
        usage_group_id2 = metrics["uuid"]["usage.profile_group_id"]

        self.assertEqual(usage_group_id1, usage_group_id2)

    def test_disabled_state_after_restart(self):
        """
        Test that the "usage-reporting" ping remains disabled and the usage ID remains null when restarting the browser.
        """

        self.enable_usage_reporting()

        # Disabling should send a "usage-deletion-request".
        ping1 = self.wait_for_ping(
            self.disable_usage_reporting,
            FOG_USAGE_DELETION_REQUEST_PING,
            ping_server=self.fog_ping_server,
        )
        metrics = ping1["payload"]["metrics"]
        self.assertIn("usage.profile_id", metrics["uuid"])
        usage_id1 = metrics["uuid"]["usage.profile_id"]
        self.assertIsValidUUID(usage_id1)
        self.assertNotEqual(CANARY_USAGE_PROFILE_ID, usage_id1)

        self.assertIn("usage.profile_group_id", metrics["uuid"])
        usage_group_id1 = metrics["uuid"]["usage.profile_group_id"]
        self.assertIsValidUUID(usage_group_id1)
        self.assertNotEqual(CANARY_USAGE_PROFILE_GROUP_ID, usage_group_id1)

        current_num_pings = len(self.fog_ping_server.pings)

        # It's not easy to wait for the _absence_ of a ping.  So restart and
        # wait for a _different_ ping, then verify we didn't get any additional
        # "usage-reporting" pings.
        _ping2 = self.wait_for_ping(
            self.restart_browser,
            BASELINE,
            ping_server=self.fog_ping_server,
        )

        self.assertIs(
            len(self.fog_ping_server.pings) > current_num_pings,
            True,
            "Expected at least 'baseline' ping",
        )
        for ping in self.fog_ping_server.pings[(current_num_pings + 1) :]:
            self.assertIs(
                FOG_USAGE_REPORTING(ping), False, "Expected no 'usage-reporting' pings"
            )

    def test_existing_profile_inherits_general_preference(self):
        # Existing profiles should inherit the general preference, true or
        # false.  New profiles should not inherit the general preference, and
        # instead should default to true.

        def healthreportEnabled():
            return self.marionette.get_pref("datareporting.healthreport.uploadEnabled")

        def usageEnabled():
            return self.marionette.get_pref("datareporting.usage.uploadEnabled")

        # Ordering matters here: we want to be sure that we really did restart the browser.
        last_pid = self.marionette.process_id

        # New profiles should not inherit the general preference.
        self.marionette.enforce_gecko_prefs(
            {"datareporting.healthreport.uploadEnabled": False}
        )
        self.assertIs(healthreportEnabled(), False)
        self.assertIs(usageEnabled(), True)
        self.assertNotEqual(last_pid, self.marionette.process_id)
        last_pid = self.marionette.process_id

        self.marionette.enforce_gecko_prefs(
            {"datareporting.healthreport.uploadEnabled": True}
        )
        self.assertIs(healthreportEnabled(), True)
        self.assertIs(usageEnabled(), True)
        self.assertNotEqual(last_pid, self.marionette.process_id)
        last_pid = self.marionette.process_id

        # Existing profiles have a migration version.
        self.marionette.enforce_gecko_prefs(
            prefs={
                "datareporting.healthreport.uploadEnabled": False,
                "browser.migration.version": 150,
            }
        )
        self.assertIs(healthreportEnabled(), False)
        self.assertIs(usageEnabled(), False)
        self.assertNotEqual(last_pid, self.marionette.process_id)
        last_pid = self.marionette.process_id

        self.marionette.enforce_gecko_prefs(
            prefs={
                "datareporting.healthreport.uploadEnabled": True,
                "browser.migration.version": 150,
            }
        )
        self.assertIs(healthreportEnabled(), True)
        self.assertIs(usageEnabled(), True)
        self.assertNotEqual(last_pid, self.marionette.process_id)
        last_pid = self.marionette.process_id



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from copy import deepcopy

from mozinfo.platforminfo import PlatformInfo
from mozunit import main

BASE_TEST_SETTINGS = {
    "platform": {
        "os": {
            "name": "linux",
            "version": "2204",
            "build": None,
        },
        "arch": "x86",
    },
    "build": {"type": "debug"},
    "runtime": {},
}


def test_os():
    test_settings = deepcopy(BASE_TEST_SETTINGS)

    # Android an linux names do not change
    test_settings["platform"]["os"]["name"] = "linux"
    test_settings["platform"]["os"]["version"] = "22.04"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os == "linux"

    test_settings["platform"]["os"]["name"] = "android"
    test_settings["platform"]["os"]["version"] = "13.0"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os == "android"

    # Macosx and Windows names are shortened
    test_settings["platform"]["os"]["name"] = "macosx"
    test_settings["platform"]["os"]["version"] = "1407"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os == "mac"

    test_settings["platform"]["os"]["name"] = "windows"
    test_settings["platform"]["os"]["version"] = "11"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os == "win"


def test_os_version():
    test_settings = deepcopy(BASE_TEST_SETTINGS)

    # linux and macosx version get expanded
    test_settings["platform"]["os"]["name"] = "linux"
    test_settings["platform"]["os"]["version"] = "2204"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os_version == "22.04"

    test_settings["platform"]["os"]["name"] = "macosx"
    test_settings["platform"]["os"]["version"] = "1470"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os_version == "14.70"

    # Macos 11 has specific hacks
    test_settings["platform"]["os"]["name"] = "macosx"
    test_settings["platform"]["os"]["version"] = "1100"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os_version == "11.20"

    # Android os version gets converted to sdk version
    test_settings["platform"]["os"]["name"] = "android"
    test_settings["platform"]["os"]["version"] = "14.0"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os_version == "34"

    # Windows version stays as is
    test_settings["platform"]["os"]["name"] = "windows"
    test_settings["platform"]["os"]["version"] = "11"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os_version == "11"

    # Can add build number if needed for windows
    test_settings["platform"]["os"]["name"] = "windows"
    test_settings["platform"]["os"]["version"] = "11"
    test_settings["platform"]["os"]["build"] = "2009"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os_version == "11.2009"

    test_settings["platform"]["os"]["name"] = "windows"
    test_settings["platform"]["os"]["version"] = "11"
    test_settings["platform"]["os"]["build"] = "24h2"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.os_version == "11.26100"


def test_os_arch():
    test_settings = deepcopy(BASE_TEST_SETTINGS)

    # detects 32bits arch as x86
    test_settings["platform"]["arch"] = "x86"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.arch == "x86"
    test_settings["platform"]["arch"] = "anything32"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.arch == "x86"

    # detects specific architectures
    test_settings["platform"]["arch"] = "aarch64"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.arch == "aarch64"
    test_settings["platform"]["arch"] = "ppc"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.arch == "ppc"
    test_settings["platform"]["arch"] = "arm7"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.arch == "arm7"

    # converts other arch as x86_64
    test_settings["platform"]["arch"] = "x86_64"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.arch == "x86_64"
    test_settings["platform"]["arch"] = "anything"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.arch == "x86_64"


def test_os_bits():
    test_settings = deepcopy(BASE_TEST_SETTINGS)

    # detects 32bits
    test_settings["platform"]["arch"] = "x86"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.bits == "32"
    test_settings["platform"]["arch"] = "anything32"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.bits == "32"

    # other architectures are assumed 64 bits
    test_settings["platform"]["arch"] = "aarch64"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.bits == "64"
    test_settings["platform"]["arch"] = "x86_64"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.bits == "64"
    test_settings["platform"]["arch"] = "anything"
    platform_info = PlatformInfo(test_settings)
    assert platform_info.bits == "64"


def test_build_type():
    test_settings = deepcopy(BASE_TEST_SETTINGS)

    # detects opt and debug build types
    test_settings["build"] = {"type": "debug"}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.build_type == "debug"
    assert platform_info.debug
    test_settings["build"] = {"type": "opt"}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.build_type == "opt"
    assert platform_info.opt

    # detects more complex build types
    test_settings["build"] = {"type": "opt", "asan": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.build_type == "asan"
    assert platform_info.opt

    # ignore shippable, devedition and mingwclang
    test_settings["build"] = {"type": "opt", "shippable": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.build_type == "opt"
    assert platform_info.opt

    test_settings["build"] = {"type": "opt", "devedition": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.build_type == "opt"
    assert platform_info.opt

    test_settings["build"] = {"type": "opt", "mingwclang": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.build_type == "opt"
    assert platform_info.opt

    # ignore ccov only on mac
    test_settings["platform"]["os"]["name"] = "macosx"
    test_settings["platform"]["os"]["version"] = "1407"
    test_settings["build"] = {"type": "opt", "ccov": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.build_type == "opt"
    assert platform_info.opt
    test_settings["platform"]["os"]["name"] = "linux"
    test_settings["platform"]["os"]["version"] = "2204"
    test_settings["build"] = {"type": "opt", "ccov": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.build_type == "ccov"
    assert platform_info.opt

    # ignore lite on android
    test_settings["platform"]["os"]["name"] = "android"
    test_settings["platform"]["os"]["version"] = "13.0"
    test_settings["build"] = {"type": "opt", "lite": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.build_type == "opt"
    assert platform_info.opt


def test_runtimes():
    test_settings = deepcopy(BASE_TEST_SETTINGS)

    # replace empty array by no_variant
    test_settings["runtime"] = {}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.test_variant == "no_variant"

    # ignore invalid runtimes
    test_settings["runtime"] = {"anything": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.test_variant == "no_variant"

    # detect valid runtimes
    test_settings["runtime"] = {"xorigin": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.test_variant == "xorigin"

    # converts variants using mowinfo
    test_settings["runtime"] = {"1proc": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.test_variant == "!e10s"

    # specific logic for no-fission
    test_settings["runtime"] = {"no-fission": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.test_variant == "!fission"

    # combines multiple runtimes
    test_settings["runtime"] = {"xorigin": True, "1proc": True}
    platform_info = PlatformInfo(test_settings)
    assert platform_info.test_variant == "xorigin+!e10s"


if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import datetime
import os
import shutil
import subprocess
import tempfile
import time

from mozdevice import ADBHost
from mozprocess import ProcessHandler

from ..errors import TimeoutException
from .base import Device
from .emulator_battery import EmulatorBattery
from .emulator_geo import EmulatorGeo
from .emulator_screen import EmulatorScreen

try:
    from telnetlib import Telnet
except ImportError:  # telnetlib was removed in Python 3.13
    from .telnetlib import Telnet


class ArchContext:
    def __init__(self, arch, context, binary=None, avd=None, extra_args=None):
        homedir = getattr(context, "homedir", "")
        kernel = os.path.join(homedir, "prebuilts", "qemu-kernel", "%s", "%s")
        sysdir = os.path.join(homedir, "out", "target", "product", "%s")
        self.extra_args = []
        self.binary = os.path.join(context.bindir or "", "emulator")
        if arch == "x86":
            self.binary = os.path.join(context.bindir or "", "emulator-x86")
            self.kernel = kernel % ("x86", "kernel-qemu")
            self.sysdir = sysdir % "generic_x86"
        elif avd:
            self.avd = avd
            self.extra_args = [
                "-show-kernel",
                "-debug",
                "init,console,gles,memcheck,adbserver,adbclient,adb,avd_config,socket",
            ]
        else:
            self.kernel = kernel % ("arm", "kernel-qemu-armv7")
            self.sysdir = sysdir % "generic"
            self.extra_args = ["-cpu", "cortex-a8"]

        if binary:
            self.binary = binary

        if extra_args:
            self.extra_args.extend(extra_args)


class SDCard:
    def __init__(self, emulator, size):
        self.emulator = emulator
        self.path = self.create_sdcard(size)

    def create_sdcard(self, sdcard_size):
        """
        Creates an sdcard partition in the emulator.

        :param sdcard_size: Size of partition to create, e.g '10MB'.
        """
        mksdcard = self.emulator.app_ctx.which("mksdcard")
        path = tempfile.mktemp(prefix="sdcard", dir=self.emulator.tmpdir)
        sdargs = [mksdcard, "-l", "mySdCard", sdcard_size, path]
        sd = subprocess.Popen(sdargs, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        retcode = sd.wait()
        if retcode:
            raise Exception(
                "unable to create sdcard: exit code %d: %s"
                % (retcode, sd.stdout.read())
            )
        return path


class BaseEmulator(Device):
    port = None
    proc = None
    telnet = None

    def __init__(self, app_ctx, **kwargs):
        self.arch = ArchContext(
            kwargs.pop("arch", "arm"),
            app_ctx,
            binary=kwargs.pop("binary", None),
            avd=kwargs.pop("avd", None),
        )
        super(BaseEmulator, self).__init__(app_ctx, **kwargs)
        self.tmpdir = tempfile.mkdtemp()
        # These rely on telnet
        self.battery = EmulatorBattery(self)
        self.geo = EmulatorGeo(self)
        self.screen = EmulatorScreen(self)

    @property
    def args(self):
        """
        Arguments to pass into the emulator binary.
        """
        return [self.arch.binary]

    def start(self):
        """
        Starts a new emulator.
        """
        if self.proc:
            return

        original_devices = set(self._get_online_devices())

        # QEMU relies on atexit() to remove temporary files, which does not
        # work since mozprocess uses SIGKILL to kill the emulator process.
        # Use a customized temporary directory so we can clean it up.
        os.environ["ANDROID_TMP"] = self.tmpdir

        qemu_log = None
        qemu_proc_args = {}
        if self.logdir:
            # save output from qemu to logfile
            qemu_log = os.path.join(self.logdir, "qemu.log")
            if os.path.isfile(qemu_log):
                self._rotate_log(qemu_log)
            qemu_proc_args["logfile"] = qemu_log
        else:
            qemu_proc_args["processOutputLine"] = lambda line: None
        self.proc = ProcessHandler(self.args, **qemu_proc_args)
        self.proc.run()

        devices = set(self._get_online_devices())
        now = datetime.datetime.now()
        while (devices - original_devices) == set([]):
            time.sleep(1)
            # Sometimes it takes more than 60s to launch emulator, so we
            # increase timeout value to 180s. Please see bug 1143380.
            if datetime.datetime.now() - now > datetime.timedelta(seconds=180):
                raise TimeoutException("timed out waiting for emulator to start")
            devices = set(self._get_online_devices())
        devices = devices - original_devices
        self.serial = devices.pop()
        self.connect()

    def _get_online_devices(self):
        adbhost = ADBHost(adb=self.app_ctx.adb)
        return [
            d["device_serial"]
            for d in adbhost.devices()
            if d["state"] != "offline"
            if d["device_serial"].startswith("emulator")
        ]

    def connect(self):
        """
        Connects to a running device. If no serial was specified in the
        constructor, defaults to the first entry in `adb devices`.
        """
        if self.connected:
            return

        super(BaseEmulator, self).connect()
        self.port = int(self.serial[self.serial.rindex("-") + 1 :])

    def cleanup(self):
        """
        Cleans up and kills the emulator, if it was started by mozrunner.
        """
        super(BaseEmulator, self).cleanup()
        if self.proc:
            self.proc.kill()
            self.proc = None
            self.connected = False

        # Remove temporary files
        if os.path.isdir(self.tmpdir):
            shutil.rmtree(self.tmpdir)

    def _get_telnet_response(self, command=None):
        output = []
        assert self.telnet
        if command is not None:
            self.telnet.write("%s\n" % command)
        while True:
            line = self.telnet.read_until("\n")
            output.append(line.rstrip())
            if line.startswith("OK"):
                return output
            elif line.startswith("KO:"):
                raise Exception("bad telnet response: %s" % line)

    def _run_telnet(self, command):
        if not self.telnet:
            self.telnet = Telnet("localhost", self.port)
            self._get_telnet_response()
        return self._get_telnet_response(command)

    def __del__(self):
        if self.telnet:
            self.telnet.write("exit\n")
            self.telnet.read_all()


class EmulatorAVD(BaseEmulator):
    def __init__(self, app_ctx, binary, avd, port=5554, **kwargs):
        super(EmulatorAVD, self).__init__(app_ctx, binary=binary, avd=avd, **kwargs)
        self.port = port

    @property
    def args(self):
        """
        Arguments to pass into the emulator binary.
        """
        qemu_args = super(EmulatorAVD, self).args
        qemu_args.extend(["-avd", self.arch.avd, "-port", str(self.port)])
        qemu_args.extend(self.arch.extra_args)
        return qemu_args

    def start(self):
        if self.proc:
            return

        env = os.environ
        env["ANDROID_AVD_HOME"] = self.app_ctx.avd_home

        super(EmulatorAVD, self).start()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


import os
import sys

# add this directory to the path
sys.path.append(os.path.dirname(__file__))
from session_store_test_case import SessionStoreTestCase


def inline(title):
    return f"data:text/html;charset=utf-8,<html><head><title>{title}</title></head><body></body></html>"


class TestTaskbarTabSessionState(SessionStoreTestCase):
    def setUp(self):
        super(TestTaskbarTabSessionState, self).setUp(
            startup_page=1,
            include_private=False,
            restore_on_demand=False,
            taskbartabs_enable=True,
            test_windows=set(
                [
                    # Window 1
                    (
                        inline("lorem ipsom"),
                        inline("dolor"),
                    ),
                ]
            ),
        )

    """
    Close all Firefox windows with the web app being closed last,
    the session store state should include the last regular window
    that's closed, but not the web app
    """

    def test_taskbartab_session_state(self):
        self.wait_for_windows(
            self.all_windows, "Not all requested windows have been opened"
        )

        self.marionette.enforce_gecko_prefs({"browser.taskbarTabs.enabled": True})

        self.open_taskbartab_window()

        # Close the original regular Firefox window
        taskbar_tab_window_handle = self.marionette.close_chrome_window()[0]
        self.marionette.switch_to_window(taskbar_tab_window_handle)

        self.marionette.set_context("content")
        dummy_html = self.marionette.absolute_url("empty.html")
        self.marionette.navigate(dummy_html)
        self.marionette.set_context("chrome")

        self.marionette.quit()
        self.marionette.start_session()
        self.marionette.set_context("chrome")

        # check for the session store state, it should have only two tabs and one window
        self.assertEqual(
            self.marionette.execute_script(
                """
                const { _LastSession } = ChromeUtils.importESModule(
                    "resource:///modules/sessionstore/SessionStore.sys.mjs"
                    );
                return _LastSession.getState().windows.length
            """
            ),
            1,
            "One window should be in the session state",
        )

        self.assertEqual(
            self.marionette.execute_script(
                """
                const { _LastSession } = ChromeUtils.importESModule(
                    "resource:///modules/sessionstore/SessionStore.sys.mjs"
                    );
                return _LastSession.getState().windows[0].tabs.length
            """
            ),
            2,
            "Two tabs should be in the session state",
        )



# mypy: allow-untyped-defs
import sys
import textwrap
from typing import Any
from typing import List
from typing import MutableSequence
from typing import NamedTuple
from typing import Optional

import attr

from _pytest import outcomes
import _pytest.assertion as plugin
from _pytest.assertion import truncate
from _pytest.assertion import util
from _pytest.config import Config as _Config
from _pytest.monkeypatch import MonkeyPatch
from _pytest.pytester import Pytester
import pytest


def mock_config(verbose: int = 0, assertion_override: Optional[int] = None):
    class TerminalWriter:
        def _highlight(self, source, lexer="python"):
            return source

    class Config:
        def get_terminal_writer(self):
            return TerminalWriter()

        def get_verbosity(self, verbosity_type: Optional[str] = None) -> int:
            if verbosity_type is None:
                return verbose
            if verbosity_type == _Config.VERBOSITY_ASSERTIONS:
                if assertion_override is not None:
                    return assertion_override
                return verbose

            raise KeyError(f"Not mocked out: {verbosity_type}")

    return Config()


class TestMockConfig:
    SOME_VERBOSITY_LEVEL = 3
    SOME_OTHER_VERBOSITY_LEVEL = 10

    def test_verbose_exposes_value(self):
        config = mock_config(verbose=TestMockConfig.SOME_VERBOSITY_LEVEL)

        assert config.get_verbosity() == TestMockConfig.SOME_VERBOSITY_LEVEL

    def test_get_assertion_override_not_set_verbose_value(self):
        config = mock_config(verbose=TestMockConfig.SOME_VERBOSITY_LEVEL)

        assert (
            config.get_verbosity(_Config.VERBOSITY_ASSERTIONS)
            == TestMockConfig.SOME_VERBOSITY_LEVEL
        )

    def test_get_assertion_override_set_custom_value(self):
        config = mock_config(
            verbose=TestMockConfig.SOME_VERBOSITY_LEVEL,
            assertion_override=TestMockConfig.SOME_OTHER_VERBOSITY_LEVEL,
        )

        assert (
            config.get_verbosity(_Config.VERBOSITY_ASSERTIONS)
            == TestMockConfig.SOME_OTHER_VERBOSITY_LEVEL
        )

    def test_get_unsupported_type_error(self):
        config = mock_config(verbose=TestMockConfig.SOME_VERBOSITY_LEVEL)

        with pytest.raises(KeyError):
            config.get_verbosity("--- NOT A VERBOSITY LEVEL ---")


class TestImportHookInstallation:
    @pytest.mark.parametrize("initial_conftest", [True, False])
    @pytest.mark.parametrize("mode", ["plain", "rewrite"])
    def test_conftest_assertion_rewrite(
        self, pytester: Pytester, initial_conftest, mode
    ) -> None:
        """Test that conftest files are using assertion rewrite on import (#1619)."""
        pytester.mkdir("foo")
        pytester.mkdir("foo/tests")
        conftest_path = "conftest.py" if initial_conftest else "foo/conftest.py"
        contents = {
            conftest_path: """
                import pytest
                @pytest.fixture
                def check_first():
                    def check(values, value):
                        assert values.pop(0) == value
                    return check
            """,
            "foo/tests/test_foo.py": """
                def test(check_first):
                    check_first([10, 30], 30)
            """,
        }
        pytester.makepyfile(**contents)
        result = pytester.runpytest_subprocess("--assert=%s" % mode)
        if mode == "plain":
            expected = "E       AssertionError"
        elif mode == "rewrite":
            expected = "*assert 10 == 30*"
        else:
            assert 0
        result.stdout.fnmatch_lines([expected])

    def test_rewrite_assertions_pytester_plugin(self, pytester: Pytester) -> None:
        """
        Assertions in the pytester plugin must also benefit from assertion
        rewriting (#1920).
        """
        pytester.makepyfile(
            """
            pytest_plugins = ['pytester']
            def test_dummy_failure(pytester):  # how meta!
                pytester.makepyfile('def test(): assert 0')
                r = pytester.inline_run()
                r.assertoutcome(passed=1)
        """
        )
        result = pytester.runpytest_subprocess()
        result.stdout.fnmatch_lines(
            [
                ">       r.assertoutcome(passed=1)",
                "E       AssertionError: ([[][]], [[][]], [[]<TestReport *>[]])*",
                "E       assert {'failed': 1,... 'skipped': 0} == {'failed': 0,... 'skipped': 0}",
                "E         Omitting 1 identical items, use -vv to show",
                "E         Differing items:",
                "E         Use -v to get more diff",
            ]
        )
        # XXX: unstable output.
        result.stdout.fnmatch_lines_random(
            [
                "E         {'failed': 1} != {'failed': 0}",
                "E         {'passed': 0} != {'passed': 1}",
            ]
        )

    @pytest.mark.parametrize("mode", ["plain", "rewrite"])
    def test_pytest_plugins_rewrite(self, pytester: Pytester, mode) -> None:
        contents = {
            "conftest.py": """
                pytest_plugins = ['ham']
            """,
            "ham.py": """
                import pytest
                @pytest.fixture
                def check_first():
                    def check(values, value):
                        assert values.pop(0) == value
                    return check
            """,
            "test_foo.py": """
                def test_foo(check_first):
                    check_first([10, 30], 30)
            """,
        }
        pytester.makepyfile(**contents)
        result = pytester.runpytest_subprocess("--assert=%s" % mode)
        if mode == "plain":
            expected = "E       AssertionError"
        elif mode == "rewrite":
            expected = "*assert 10 == 30*"
        else:
            assert 0
        result.stdout.fnmatch_lines([expected])

    @pytest.mark.parametrize("mode", ["str", "list"])
    def test_pytest_plugins_rewrite_module_names(
        self, pytester: Pytester, mode
    ) -> None:
        """Test that pluginmanager correct marks pytest_plugins variables
        for assertion rewriting if they are defined as plain strings or
        list of strings (#1888).
        """
        plugins = '"ham"' if mode == "str" else '["ham"]'
        contents = {
            "conftest.py": f"""
                pytest_plugins = {plugins}
            """,
            "ham.py": """
                import pytest
            """,
            "test_foo.py": """
                def test_foo(pytestconfig):
                    assert 'ham' in pytestconfig.pluginmanager.rewrite_hook._must_rewrite
            """,
        }
        pytester.makepyfile(**contents)
        result = pytester.runpytest_subprocess("--assert=rewrite")
        assert result.ret == 0

    def test_pytest_plugins_rewrite_module_names_correctly(
        self, pytester: Pytester
    ) -> None:
        """Test that we match files correctly when they are marked for rewriting (#2939)."""
        contents = {
            "conftest.py": """\
                pytest_plugins = "ham"
            """,
            "ham.py": "",
            "hamster.py": "",
            "test_foo.py": """\
                def test_foo(pytestconfig):
                    assert pytestconfig.pluginmanager.rewrite_hook.find_spec('ham') is not None
                    assert pytestconfig.pluginmanager.rewrite_hook.find_spec('hamster') is None
            """,
        }
        pytester.makepyfile(**contents)
        result = pytester.runpytest_subprocess("--assert=rewrite")
        assert result.ret == 0

    @pytest.mark.parametrize("mode", ["plain", "rewrite"])
    def test_installed_plugin_rewrite(
        self, pytester: Pytester, mode, monkeypatch
    ) -> None:
        monkeypatch.delenv("PYTEST_DISABLE_PLUGIN_AUTOLOAD", raising=False)
        # Make sure the hook is installed early enough so that plugins
        # installed via setuptools are rewritten.
        pytester.mkdir("hampkg")
        contents = {
            "hampkg/__init__.py": """\
                import pytest

                @pytest.fixture
                def check_first2():
                    def check(values, value):
                        assert values.pop(0) == value
                    return check
            """,
            "spamplugin.py": """\
            import pytest
            from hampkg import check_first2

            @pytest.fixture
            def check_first():
                def check(values, value):
                    assert values.pop(0) == value
                return check
            """,
            "mainwrapper.py": """\
            import importlib.metadata
            import pytest

            class DummyEntryPoint(object):
                name = 'spam'
                module_name = 'spam.py'
                group = 'pytest11'

                def load(self):
                    import spamplugin
                    return spamplugin

            class DummyDistInfo(object):
                version = '1.0'
                files = ('spamplugin.py', 'hampkg/__init__.py')
                entry_points = (DummyEntryPoint(),)
                metadata = {'name': 'foo'}

            def distributions():
                return (DummyDistInfo(),)

            importlib.metadata.distributions = distributions
            pytest.main()
            """,
            "test_foo.py": """\
            def test(check_first):
                check_first([10, 30], 30)

            def test2(check_first2):
                check_first([10, 30], 30)
            """,
        }
        pytester.makepyfile(**contents)
        result = pytester.run(
            sys.executable, "mainwrapper.py", "-s", "--assert=%s" % mode
        )
        if mode == "plain":
            expected = "E       AssertionError"
        elif mode == "rewrite":
            expected = "*assert 10 == 30*"
        else:
            assert 0
        result.stdout.fnmatch_lines([expected])

    def test_rewrite_ast(self, pytester: Pytester) -> None:
        pytester.mkdir("pkg")
        contents = {
            "pkg/__init__.py": """
                import pytest
                pytest.register_assert_rewrite('pkg.helper')
            """,
            "pkg/helper.py": """
                def tool():
                    a, b = 2, 3
                    assert a == b
            """,
            "pkg/plugin.py": """
                import pytest, pkg.helper
                @pytest.fixture
                def tool():
                    return pkg.helper.tool
            """,
            "pkg/other.py": """
                values = [3, 2]
                def tool():
                    assert values.pop() == 3
            """,
            "conftest.py": """
                pytest_plugins = ['pkg.plugin']
            """,
            "test_pkg.py": """
                import pkg.other
                def test_tool(tool):
                    tool()
                def test_other():
                    pkg.other.tool()
            """,
        }
        pytester.makepyfile(**contents)
        result = pytester.runpytest_subprocess("--assert=rewrite")
        result.stdout.fnmatch_lines(
            [
                ">*assert a == b*",
                "E*assert 2 == 3*",
                ">*assert values.pop() == 3*",
                "E*AssertionError",
            ]
        )

    def test_register_assert_rewrite_checks_types(self) -> None:
        with pytest.raises(TypeError):
            pytest.register_assert_rewrite(["pytest_tests_internal_non_existing"])  # type: ignore
        pytest.register_assert_rewrite(
            "pytest_tests_internal_non_existing", "pytest_tests_internal_non_existing2"
        )


class TestBinReprIntegration:
    def test_pytest_assertrepr_compare_called(self, pytester: Pytester) -> None:
        pytester.makeconftest(
            """
            import pytest
            values = []
            def pytest_assertrepr_compare(op, left, right):
                values.append((op, left, right))

            @pytest.fixture
            def list(request):
                return values
        """
        )
        pytester.makepyfile(
            """
            def test_hello():
                assert 0 == 1
            def test_check(list):
                assert list == [("==", 0, 1)]
        """
        )
        result = pytester.runpytest("-v")
        result.stdout.fnmatch_lines(["*test_hello*FAIL*", "*test_check*PASS*"])


def callop(op: str, left: Any, right: Any, verbose: int = 0) -> Optional[List[str]]:
    config = mock_config(verbose=verbose)
    return plugin.pytest_assertrepr_compare(config, op, left, right)


def callequal(left: Any, right: Any, verbose: int = 0) -> Optional[List[str]]:
    return callop("==", left, right, verbose)


class TestAssert_reprcompare:
    def test_different_types(self) -> None:
        assert callequal([0, 1], "foo") is None

    def test_summary(self) -> None:
        lines = callequal([0, 1], [0, 2])
        assert lines is not None
        summary = lines[0]
        assert len(summary) < 65

    def test_text_diff(self) -> None:
        assert callequal("spam", "eggs") == [
            "'spam' == 'eggs'",
            "",
            "- eggs",
            "+ spam",
        ]

    def test_text_skipping(self) -> None:
        lines = callequal("a" * 50 + "spam", "a" * 50 + "eggs")
        assert lines is not None
        assert "Skipping" in lines[2]
        for line in lines:
            assert "a" * 50 not in line

    def test_text_skipping_verbose(self) -> None:
        lines = callequal("a" * 50 + "spam", "a" * 50 + "eggs", verbose=1)
        assert lines is not None
        assert "- " + "a" * 50 + "eggs" in lines
        assert "+ " + "a" * 50 + "spam" in lines

    def test_multiline_text_diff(self) -> None:
        left = "foo\nspam\nbar"
        right = "foo\neggs\nbar"
        diff = callequal(left, right)
        assert diff is not None
        assert "- eggs" in diff
        assert "+ spam" in diff

    def test_bytes_diff_normal(self) -> None:
        """Check special handling for bytes diff (#5260)"""
        diff = callequal(b"spam", b"eggs")

        assert diff == [
            "b'spam' == b'eggs'",
            "",
            "At index 0 diff: b's' != b'e'",
            "Use -v to get more diff",
        ]

    def test_bytes_diff_verbose(self) -> None:
        """Check special handling for bytes diff (#5260)"""
        diff = callequal(b"spam", b"eggs", verbose=1)
        assert diff == [
            "b'spam' == b'eggs'",
            "",
            "At index 0 diff: b's' != b'e'",
            "",
            "Full diff:",
            "- b'eggs'",
            "+ b'spam'",
        ]

    def test_list(self) -> None:
        expl = callequal([0, 1], [0, 2])
        assert expl is not None
        assert len(expl) > 1

    @pytest.mark.parametrize(
        ["left", "right", "expected"],
        [
            pytest.param(
                [0, 1],
                [0, 2],
                """
                Full diff:
                  [
                      0,
                -     2,
                ?     ^
                +     1,
                ?     ^
                  ]
                """,
                id="lists",
            ),
            pytest.param(
                {0: 1},
                {0: 2},
                """
                Full diff:
                  {
                -     0: 2,
                ?        ^
                +     0: 1,
                ?        ^
                  }
            """,
                id="dicts",
            ),
            pytest.param(
                {0, 1},
                {0, 2},
                """
                Full diff:
                  {
                      0,
                -     2,
                ?     ^
                +     1,
                ?     ^
                  }
            """,
                id="sets",
            ),
        ],
    )
    def test_iterable_full_diff(self, left, right, expected) -> None:
        """Test the full diff assertion failure explanation.

        When verbose is False, then just a -v notice to get the diff is rendered,
        when verbose is True, then ndiff of the pprint is returned.
        """
        expl = callequal(left, right, verbose=0)
        assert expl is not None
        assert expl[-1] == "Use -v to get more diff"
        verbose_expl = callequal(left, right, verbose=1)
        assert verbose_expl is not None
        assert "\n".join(verbose_expl).endswith(textwrap.dedent(expected).strip())

    def test_iterable_quiet(self) -> None:
        expl = callequal([1, 2], [10, 2], verbose=-1)
        assert expl == [
            "[1, 2] == [10, 2]",
            "",
            "At index 0 diff: 1 != 10",
            "Use -v to get more diff",
        ]

    def test_iterable_full_diff_ci(
        self, monkeypatch: MonkeyPatch, pytester: Pytester
    ) -> None:
        pytester.makepyfile(
            r"""
            def test_full_diff():
                left = [0, 1]
                right = [0, 2]
                assert left == right
        """
        )
        monkeypatch.setenv("CI", "true")
        result = pytester.runpytest()
        result.stdout.fnmatch_lines(["E         Full diff:"])

        monkeypatch.delenv("CI", raising=False)
        result = pytester.runpytest()
        result.stdout.fnmatch_lines(["E         Use -v to get more diff"])

    def test_list_different_lengths(self) -> None:
        expl = callequal([0, 1], [0, 1, 2])
        assert expl is not None
        assert len(expl) > 1
        expl = callequal([0, 1, 2], [0, 1])
        assert expl is not None
        assert len(expl) > 1

    def test_list_wrap_for_multiple_lines(self) -> None:
        long_d = "d" * 80
        l1 = ["a", "b", "c"]
        l2 = ["a", "b", "c", long_d]
        diff = callequal(l1, l2, verbose=True)
        assert diff == [
            "['a', 'b', 'c'] == ['a', 'b', 'c...dddddddddddd']",
            "",
            "Right contains one more item: '" + long_d + "'",
            "",
            "Full diff:",
            "  [",
            "      'a',",
            "      'b',",
            "      'c',",
            "-     '" + long_d + "',",
            "  ]",
        ]

        diff = callequal(l2, l1, verbose=True)
        assert diff == [
            "['a', 'b', 'c...dddddddddddd'] == ['a', 'b', 'c']",
            "",
            "Left contains one more item: '" + long_d + "'",
            "",
            "Full diff:",
            "  [",
            "      'a',",
            "      'b',",
            "      'c',",
            "+     '" + long_d + "',",
            "  ]",
        ]

    def test_list_wrap_for_width_rewrap_same_length(self) -> None:
        long_a = "a" * 30
        long_b = "b" * 30
        long_c = "c" * 30
        l1 = [long_a, long_b, long_c]
        l2 = [long_b, long_c, long_a]
        diff = callequal(l1, l2, verbose=True)
        assert diff == [
            "['aaaaaaaaaaa...cccccccccccc'] == ['bbbbbbbbbbb...aaaaaaaaaaaa']",
            "",
            "At index 0 diff: 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' != 'bbbbbbbbbbbbbbbbbbbbbbbbbbbbbb'",
            "",
            "Full diff:",
            "  [",
            "+     'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',",
            "      'bbbbbbbbbbbbbbbbbbbbbbbbbbbbbb',",
            "      'cccccccccccccccccccccccccccccc',",
            "-     'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa',",
            "  ]",
        ]

    def test_list_dont_wrap_strings(self) -> None:
        long_a = "a" * 10
        l1 = ["a"] + [long_a for _ in range(7)]
        l2 = ["should not get wrapped"]
        diff = callequal(l1, l2, verbose=True)
        assert diff == [
            "['a', 'aaaaaa...aaaaaaa', ...] == ['should not get wrapped']",
            "",
            "At index 0 diff: 'a' != 'should not get wrapped'",
            "Left contains 7 more items, first extra item: 'aaaaaaaaaa'",
            "",
            "Full diff:",
            "  [",
            "-     'should not get wrapped',",
            "+     'a',",
            "+     'aaaaaaaaaa',",
            "+     'aaaaaaaaaa',",
            "+     'aaaaaaaaaa',",
            "+     'aaaaaaaaaa',",
            "+     'aaaaaaaaaa',",
            "+     'aaaaaaaaaa',",
            "+     'aaaaaaaaaa',",
            "  ]",
        ]

    def test_dict_wrap(self) -> None:
        d1 = {"common": 1, "env": {"env1": 1, "env2": 2}}
        d2 = {"common": 1, "env": {"env1": 1}}

        diff = callequal(d1, d2, verbose=True)
        assert diff == [
            "{'common': 1,...1, 'env2': 2}} == {'common': 1,...: {'env1': 1}}",
            "",
            "Omitting 1 identical items, use -vv to show",
            "Differing items:",
            "{'env': {'env1': 1, 'env2': 2}} != {'env': {'env1': 1}}",
            "",
            "Full diff:",
            "  {",
            "      'common': 1,",
            "      'env': {",
            "          'env1': 1,",
            "+         'env2': 2,",
            "      },",
            "  }",
        ]

        long_a = "a" * 80
        sub = {"long_a": long_a, "sub1": {"long_a": "substring that gets wrapped " * 3}}
        d1 = {"env": {"sub": sub}}
        d2 = {"env": {"sub": sub}, "new": 1}
        diff = callequal(d1, d2, verbose=True)
        assert diff == [
            "{'env': {'sub... wrapped '}}}} == {'env': {'sub...}}}, 'new': 1}",
            "",
            "Omitting 1 identical items, use -vv to show",
            "Right contains 1 more item:",
            "{'new': 1}",
            "",
            "Full diff:",
            "  {",
            "      'env': {",
            "          'sub': {",
            f"              'long_a': '{long_a}',",
            "              'sub1': {",
            "                  'long_a': 'substring that gets wrapped substring that gets wrapped '",
            "                  'substring that gets wrapped ',",
            "              },",
            "          },",
            "      },",
            "-     'new': 1,",
            "  }",
        ]

    def test_dict(self) -> None:
        expl = callequal({"a": 0}, {"a": 1})
        assert expl is not None
        assert len(expl) > 1

    def test_dict_omitting(self) -> None:
        lines = callequal({"a": 0, "b": 1}, {"a": 1, "b": 1})
        assert lines is not None
        assert lines[2].startswith("Omitting 1 identical item")
        assert "Common items" not in lines
        for line in lines[1:]:
            assert "b" not in line

    def test_dict_omitting_with_verbosity_1(self) -> None:
        """Ensure differing items are visible for verbosity=1 (#1512)."""
        lines = callequal({"a": 0, "b": 1}, {"a": 1, "b": 1}, verbose=1)
        assert lines is not None
        assert lines[1] == ""
        assert lines[2].startswith("Omitting 1 identical item")
        assert lines[3].startswith("Differing items")
        assert lines[4] == "{'a': 0} != {'a': 1}"
        assert "Common items" not in lines

    def test_dict_omitting_with_verbosity_2(self) -> None:
        lines = callequal({"a": 0, "b": 1}, {"a": 1, "b": 1}, verbose=2)
        assert lines is not None
        assert lines[2].startswith("Common items:")
        assert "Omitting" not in lines[2]
        assert lines[3] == "{'b': 1}"

    def test_dict_different_items(self) -> None:
        lines = callequal({"a": 0}, {"b": 1, "c": 2}, verbose=2)
        assert lines == [
            "{'a': 0} == {'b': 1, 'c': 2}",
            "",
            "Left contains 1 more item:",
            "{'a': 0}",
            "Right contains 2 more items:",
            "{'b': 1, 'c': 2}",
            "",
            "Full diff:",
            "  {",
            "-     'b': 1,",
            "?      ^   ^",
            "+     'a': 0,",
            "?      ^   ^",
            "-     'c': 2,",
            "  }",
        ]
        lines = callequal({"b": 1, "c": 2}, {"a": 0}, verbose=2)
        assert lines == [
            "{'b': 1, 'c': 2} == {'a': 0}",
            "",
            "Left contains 2 more items:",
            "{'b': 1, 'c': 2}",
            "Right contains 1 more item:",
            "{'a': 0}",
            "",
            "Full diff:",
            "  {",
            "-     'a': 0,",
            "?      ^   ^",
            "+     'b': 1,",
            "?      ^   ^",
            "+     'c': 2,",
            "  }",
        ]

    def test_sequence_different_items(self) -> None:
        lines = callequal((1, 2), (3, 4, 5), verbose=2)
        assert lines == [
            "(1, 2) == (3, 4, 5)",
            "",
            "At index 0 diff: 1 != 3",
            "Right contains one more item: 5",
            "",
            "Full diff:",
            "  (",
            "-     3,",
            "?     ^",
            "+     1,",
            "?     ^",
            "-     4,",
            "?     ^",
            "+     2,",
            "?     ^",
            "-     5,",
            "  )",
        ]
        lines = callequal((1, 2, 3), (4,), verbose=2)
        assert lines == [
            "(1, 2, 3) == (4,)",
            "",
            "At index 0 diff: 1 != 4",
            "Left contains 2 more items, first extra item: 2",
            "",
            "Full diff:",
            "  (",
            "-     4,",
            "?     ^",
            "+     1,",
            "?     ^",
            "+     2,",
            "+     3,",
            "  )",
        ]
        lines = callequal((1, 2, 3), (1, 20, 3), verbose=2)
        assert lines == [
            "(1, 2, 3) == (1, 20, 3)",
            "",
            "At index 1 diff: 2 != 20",
            "",
            "Full diff:",
            "  (",
            "      1,",
            "-     20,",
            "?      -",
            "+     2,",
            "      3,",
            "  )",
        ]

    def test_set(self) -> None:
        expl = callequal({0, 1}, {0, 2})
        assert expl is not None
        assert len(expl) > 1

    def test_frozenzet(self) -> None:
        expl = callequal(frozenset([0, 1]), {0, 2})
        assert expl is not None
        assert len(expl) > 1

    def test_Sequence(self) -> None:
        # Test comparing with a Sequence subclass.
        class TestSequence(MutableSequence[int]):
            def __init__(self, iterable):
                self.elements = list(iterable)

            def __getitem__(self, item):
                return self.elements[item]

            def __len__(self):
                return len(self.elements)

            def __setitem__(self, item, value):
                pass

            def __delitem__(self, item):
                pass

            def insert(self, item, index):
                pass

        expl = callequal(TestSequence([0, 1]), list([0, 2]))
        assert expl is not None
        assert len(expl) > 1

    def test_list_tuples(self) -> None:
        expl = callequal([], [(1, 2)])
        assert expl is not None
        assert len(expl) > 1
        expl = callequal([(1, 2)], [])
        assert expl is not None
        assert len(expl) > 1

    def test_list_bad_repr(self) -> None:
        class A:
            def __repr__(self):
                raise ValueError(42)

        expl = callequal([], [A()])
        assert expl is not None
        assert "ValueError" in "".join(expl)
        expl = callequal({}, {"1": A()}, verbose=2)
        assert expl is not None
        assert expl[0].startswith("{} == <[ValueError")
        assert "raised in repr" in expl[0]
        assert expl[2:] == [
            "(pytest_assertion plugin: representation of details failed:"
            f" {__file__}:{A.__repr__.__code__.co_firstlineno + 1}: ValueError: 42.",
            " Probably an object has a faulty __repr__.)",
        ]

    def test_one_repr_empty(self) -> None:
        """The faulty empty string repr did trigger an unbound local error in _diff_text."""

        class A(str):
            def __repr__(self):
                return ""

        expl = callequal(A(), "")
        assert not expl

    def test_repr_no_exc(self) -> None:
        expl = callequal("foo", "bar")
        assert expl is not None
        assert "raised in repr()" not in " ".join(expl)

    def test_unicode(self) -> None:
        assert callequal("£€", "£") == [
            "'£€' == '£'",
            "",
            "- £",
            "+ £€",
        ]

    def test_nonascii_text(self) -> None:
        """
        :issue: 877
        non ascii python2 str caused a UnicodeDecodeError
        """

        class A(str):
            def __repr__(self):
                return "\xff"

        expl = callequal(A(), "1")
        assert expl == ["ÿ == '1'", "", "- 1"]

    def test_format_nonascii_explanation(self) -> None:
        assert util.format_explanation("λ")

    def test_mojibake(self) -> None:
        # issue 429
        left = b"e"
        right = b"\xc3\xa9"
        expl = callequal(left, right)
        assert expl is not None
        for line in expl:
            assert isinstance(line, str)
        msg = "\n".join(expl)
        assert msg

    def test_nfc_nfd_same_string(self) -> None:
        # issue 3426
        left = "hyv\xe4"
        right = "hyva\u0308"
        expl = callequal(left, right)
        assert expl == [
            r"'hyv\xe4' == 'hyva\u0308'",
            "",
            f"- {right!s}",
            f"+ {left!s}",
        ]

        expl = callequal(left, right, verbose=2)
        assert expl == [
            r"'hyv\xe4' == 'hyva\u0308'",
            "",
            f"- {right!s}",
            f"+ {left!s}",
        ]


class TestAssert_reprcompare_dataclass:
    def test_dataclasses(self, pytester: Pytester) -> None:
        p = pytester.copy_example("dataclasses/test_compare_dataclasses.py")
        result = pytester.runpytest(p)
        result.assert_outcomes(failed=1, passed=0)
        result.stdout.fnmatch_lines(
            [
                "E         Omitting 1 identical items, use -vv to show",
                "E         Differing attributes:",
                "E         ['field_b']",
                "E         ",
                "E         Drill down into differing attribute field_b:",
                "E           field_b: 'b' != 'c'",
                "E           - c",
                "E           + b",
            ],
            consecutive=True,
        )

    def test_recursive_dataclasses(self, pytester: Pytester) -> None:
        p = pytester.copy_example("dataclasses/test_compare_recursive_dataclasses.py")
        result = pytester.runpytest(p)
        result.assert_outcomes(failed=1, passed=0)
        result.stdout.fnmatch_lines(
            [
                "E         Omitting 1 identical items, use -vv to show",
                "E         Differing attributes:",
                "E         ['g', 'h', 'j']",
                "E         ",
                "E         Drill down into differing attribute g:",
                "E           g: S(a=10, b='ten') != S(a=20, b='xxx')...",
                "E         ",
                "E         ...Full output truncated (51 lines hidden), use '-vv' to show",
            ],
            consecutive=True,
        )

    def test_recursive_dataclasses_verbose(self, pytester: Pytester) -> None:
        p = pytester.copy_example("dataclasses/test_compare_recursive_dataclasses.py")
        result = pytester.runpytest(p, "-vv")
        result.assert_outcomes(failed=1, passed=0)
        result.stdout.fnmatch_lines(
            [
                "E         Matching attributes:",
                "E         ['i']",
                "E         Differing attributes:",
                "E         ['g', 'h', 'j']",
                "E         ",
                "E         Drill down into differing attribute g:",
                "E           g: S(a=10, b='ten') != S(a=20, b='xxx')",
                "E           ",
                "E           Differing attributes:",
                "E           ['a', 'b']",
                "E           ",
                "E           Drill down into differing attribute a:",
                "E             a: 10 != 20",
                "E           ",
                "E           Drill down into differing attribute b:",
                "E             b: 'ten' != 'xxx'",
                "E             - xxx",
                "E             + ten",
                "E         ",
                "E         Drill down into differing attribute h:",
            ],
            consecutive=True,
        )

    def test_dataclasses_verbose(self, pytester: Pytester) -> None:
        p = pytester.copy_example("dataclasses/test_compare_dataclasses_verbose.py")
        result = pytester.runpytest(p, "-vv")
        result.assert_outcomes(failed=1, passed=0)
        result.stdout.fnmatch_lines(
            [
                "*Matching attributes:*",
                "*['field_a']*",
                "*Differing attributes:*",
                "*field_b: 'b' != 'c'*",
            ]
        )

    def test_dataclasses_with_attribute_comparison_off(
        self, pytester: Pytester
    ) -> None:
        p = pytester.copy_example(
            "dataclasses/test_compare_dataclasses_field_comparison_off.py"
        )
        result = pytester.runpytest(p, "-vv")
        result.assert_outcomes(failed=0, passed=1)

    def test_comparing_two_different_data_classes(self, pytester: Pytester) -> None:
        p = pytester.copy_example(
            "dataclasses/test_compare_two_different_dataclasses.py"
        )
        result = pytester.runpytest(p, "-vv")
        result.assert_outcomes(failed=0, passed=1)

    def test_data_classes_with_custom_eq(self, pytester: Pytester) -> None:
        p = pytester.copy_example(
            "dataclasses/test_compare_dataclasses_with_custom_eq.py"
        )
        # issue 9362
        result = pytester.runpytest(p, "-vv")
        result.assert_outcomes(failed=1, passed=0)
        result.stdout.no_re_match_line(".*Differing attributes.*")

    def test_data_classes_with_initvar(self, pytester: Pytester) -> None:
        p = pytester.copy_example("dataclasses/test_compare_initvar.py")
        # issue 9820
        result = pytester.runpytest(p, "-vv")
        result.assert_outcomes(failed=1, passed=0)
        result.stdout.no_re_match_line(".*AttributeError.*")


class TestAssert_reprcompare_attrsclass:
    def test_attrs(self) -> None:
        @attr.s
        class SimpleDataObject:
            field_a = attr.ib()
            field_b = attr.ib()

        left = SimpleDataObject(1, "b")
        right = SimpleDataObject(1, "c")

        lines = callequal(left, right)
        assert lines is not None
        assert lines[2].startswith("Omitting 1 identical item")
        assert "Matching attributes" not in lines
        for line in lines[2:]:
            assert "field_a" not in line

    def test_attrs_recursive(self) -> None:
        @attr.s
        class OtherDataObject:
            field_c = attr.ib()
            field_d = attr.ib()

        @attr.s
        class SimpleDataObject:
            field_a = attr.ib()
            field_b = attr.ib()

        left = SimpleDataObject(OtherDataObject(1, "a"), "b")
        right = SimpleDataObject(OtherDataObject(1, "b"), "b")

        lines = callequal(left, right)
        assert lines is not None
        assert "Matching attributes" not in lines
        for line in lines[1:]:
            assert "field_b:" not in line
            assert "field_c:" not in line

    def test_attrs_recursive_verbose(self) -> None:
        @attr.s
        class OtherDataObject:
            field_c = attr.ib()
            field_d = attr.ib()

        @attr.s
        class SimpleDataObject:
            field_a = attr.ib()
            field_b = attr.ib()

        left = SimpleDataObject(OtherDataObject(1, "a"), "b")
        right = SimpleDataObject(OtherDataObject(1, "b"), "b")

        lines = callequal(left, right)
        assert lines is not None
        # indentation in output because of nested object structure
        assert "    field_d: 'a' != 'b'" in lines

    def test_attrs_verbose(self) -> None:
        @attr.s
        class SimpleDataObject:
            field_a = attr.ib()
            field_b = attr.ib()

        left = SimpleDataObject(1, "b")
        right = SimpleDataObject(1, "c")

        lines = callequal(left, right, verbose=2)
        assert lines is not None
        assert lines[2].startswith("Matching attributes:")
        assert "Omitting" not in lines[2]
        assert lines[3] == "['field_a']"

    def test_attrs_with_attribute_comparison_off(self) -> None:
        @attr.s
        class SimpleDataObject:
            field_a = attr.ib()
            field_b = attr.ib(eq=False)

        left = SimpleDataObject(1, "b")
        right = SimpleDataObject(1, "b")

        lines = callequal(left, right, verbose=2)
        assert lines is not None
        assert lines[2].startswith("Matching attributes:")
        assert "Omitting" not in lines[1]
        assert lines[3] == "['field_a']"
        for line in lines[3:]:
            assert "field_b" not in line

    def test_comparing_two_different_attrs_classes(self) -> None:
        @attr.s
        class SimpleDataObjectOne:
            field_a = attr.ib()
            field_b = attr.ib()

        @attr.s
        class SimpleDataObjectTwo:
            field_a = attr.ib()
            field_b = attr.ib()

        left = SimpleDataObjectOne(1, "b")
        right = SimpleDataObjectTwo(1, "c")

        lines = callequal(left, right)
        assert lines is None

    def test_attrs_with_auto_detect_and_custom_eq(self) -> None:
        @attr.s(
            auto_detect=True
        )  # attr.s doesn't ignore a custom eq if auto_detect=True
        class SimpleDataObject:
            field_a = attr.ib()

            def __eq__(self, other):  # pragma: no cover
                return super().__eq__(other)

        left = SimpleDataObject(1)
        right = SimpleDataObject(2)
        # issue 9362
        lines = callequal(left, right, verbose=2)
        assert lines is None

    def test_attrs_with_custom_eq(self) -> None:
        @attr.define(slots=False)
        class SimpleDataObject:
            field_a = attr.ib()

            def __eq__(self, other):  # pragma: no cover
                return super().__eq__(other)

        left = SimpleDataObject(1)
        right = SimpleDataObject(2)
        # issue 9362
        lines = callequal(left, right, verbose=2)
        assert lines is None


class TestAssert_reprcompare_namedtuple:
    def test_namedtuple(self) -> None:
        class NT(NamedTuple):
            a: Any
            b: Any

        left = NT(1, "b")
        right = NT(1, "c")

        lines = callequal(left, right)
        assert lines == [
            "NT(a=1, b='b') == NT(a=1, b='c')",
            "",
            "Omitting 1 identical items, use -vv to show",
            "Differing attributes:",
            "['b']",
            "",
            "Drill down into differing attribute b:",
            "  b: 'b' != 'c'",
            "  - c",
            "  + b",
            "Use -v to get more diff",
        ]

    def test_comparing_two_different_namedtuple(self) -> None:
        class NT1(NamedTuple):
            a: Any
            b: Any

        class NT2(NamedTuple):
            a: Any
            b: Any

        left = NT1(1, "b")
        right = NT2(2, "b")

        lines = callequal(left, right)
        # Because the types are different, uses the generic sequence matcher.
        assert lines == [
            "NT1(a=1, b='b') == NT2(a=2, b='b')",
            "",
            "At index 0 diff: 1 != 2",
            "Use -v to get more diff",
        ]


class TestFormatExplanation:
    def test_special_chars_full(self, pytester: Pytester) -> None:
        # Issue 453, for the bug this would raise IndexError
        pytester.makepyfile(
            """
            def test_foo():
                assert '\\n}' == ''
        """
        )
        result = pytester.runpytest()
        assert result.ret == 1
        result.stdout.fnmatch_lines(["*AssertionError*"])

    def test_fmt_simple(self) -> None:
        expl = "assert foo"
        assert util.format_explanation(expl) == "assert foo"

    def test_fmt_where(self) -> None:
        expl = "\n".join(["assert 1", "{1 = foo", "} == 2"])
        res = "\n".join(["assert 1 == 2", " +  where 1 = foo"])
        assert util.format_explanation(expl) == res

    def test_fmt_and(self) -> None:
        expl = "\n".join(["assert 1", "{1 = foo", "} == 2", "{2 = bar", "}"])
        res = "\n".join(["assert 1 == 2", " +  where 1 = foo", " +  and   2 = bar"])
        assert util.format_explanation(expl) == res

    def test_fmt_where_nested(self) -> None:
        expl = "\n".join(["assert 1", "{1 = foo", "{foo = bar", "}", "} == 2"])
        res = "\n".join(["assert 1 == 2", " +  where 1 = foo", " +    where foo = bar"])
        assert util.format_explanation(expl) == res

    def test_fmt_newline(self) -> None:
        expl = "\n".join(['assert "foo" == "bar"', "~- foo", "~+ bar"])
        res = "\n".join(['assert "foo" == "bar"', "  - foo", "  + bar"])
        assert util.format_explanation(expl) == res

    def test_fmt_newline_escaped(self) -> None:
        expl = "\n".join(["assert foo == bar", "baz"])
        res = "assert foo == bar\\nbaz"
        assert util.format_explanation(expl) == res

    def test_fmt_newline_before_where(self) -> None:
        expl = "\n".join(
            [
                "the assertion message here",
                ">assert 1",
                "{1 = foo",
                "} == 2",
                "{2 = bar",
                "}",
            ]
        )
        res = "\n".join(
            [
                "the assertion message here",
                "assert 1 == 2",
                " +  where 1 = foo",
                " +  and   2 = bar",
            ]
        )
        assert util.format_explanation(expl) == res

    def test_fmt_multi_newline_before_where(self) -> None:
        expl = "\n".join(
            [
                "the assertion",
                "~message here",
                ">assert 1",
                "{1 = foo",
                "} == 2",
                "{2 = bar",
                "}",
            ]
        )
        res = "\n".join(
            [
                "the assertion",
                "  message here",
                "assert 1 == 2",
                " +  where 1 = foo",
                " +  and   2 = bar",
            ]
        )
        assert util.format_explanation(expl) == res


class TestTruncateExplanation:
    # The number of lines in the truncation explanation message. Used
    # to calculate that results have the expected length.
    LINES_IN_TRUNCATION_MSG = 2

    def test_doesnt_truncate_when_input_is_empty_list(self) -> None:
        expl: List[str] = []
        result = truncate._truncate_explanation(expl, max_lines=8, max_chars=100)
        assert result == expl

    def test_doesnt_truncate_at_when_input_is_5_lines_and_LT_max_chars(self) -> None:
        expl = ["a" * 100 for x in range(5)]
        result = truncate._truncate_explanation(expl, max_lines=8, max_chars=8 * 80)
        assert result == expl

    def test_truncates_at_8_lines_when_given_list_of_empty_strings(self) -> None:
        expl = ["" for x in range(50)]
        result = truncate._truncate_explanation(expl, max_lines=8, max_chars=100)
        assert len(result) != len(expl)
        assert result != expl
        assert len(result) == 8 + self.LINES_IN_TRUNCATION_MSG
        assert "Full output truncated" in result[-1]
        assert "42 lines hidden" in result[-1]
        last_line_before_trunc_msg = result[-self.LINES_IN_TRUNCATION_MSG - 1]
        assert last_line_before_trunc_msg.endswith("...")

    def test_truncates_at_8_lines_when_first_8_lines_are_LT_max_chars(self) -> None:
        total_lines = 100
        expl = ["a" for x in range(total_lines)]
        result = truncate._truncate_explanation(expl, max_lines=8, max_chars=8 * 80)
        assert result != expl
        assert len(result) == 8 + self.LINES_IN_TRUNCATION_MSG
        assert "Full output truncated" in result[-1]
        assert f"{total_lines - 8} lines hidden" in result[-1]
        last_line_before_trunc_msg = result[-self.LINES_IN_TRUNCATION_MSG - 1]
        assert last_line_before_trunc_msg.endswith("...")

    def test_truncates_at_8_lines_when_there_is_one_line_to_remove(self) -> None:
        """The number of line in the result is 9, the same number as if we truncated."""
        expl = ["a" for x in range(9)]
        result = truncate._truncate_explanation(expl, max_lines=8, max_chars=8 * 80)
        assert result == expl
        assert "truncated" not in result[-1]

    def test_truncates_edgecase_when_truncation_message_makes_the_result_longer_for_chars(
        self,
    ) -> None:
        line = "a" * 10
        expl = [line, line]
        result = truncate._truncate_explanation(expl, max_lines=10, max_chars=10)
        assert result == [line, line]

    def test_truncates_edgecase_when_truncation_message_makes_the_result_longer_for_lines(
        self,
    ) -> None:
        line = "a" * 10
        expl = [line, line]
        result = truncate._truncate_explanation(expl, max_lines=1, max_chars=100)
        assert result == [line, line]

    def test_truncates_at_8_lines_when_first_8_lines_are_EQ_max_chars(self) -> None:
        expl = [chr(97 + x) * 80 for x in range(16)]
        result = truncate._truncate_explanation(expl, max_lines=8, max_chars=8 * 80)
        assert result != expl
        assert len(result) == 16 - 8 + self.LINES_IN_TRUNCATION_MSG
        assert "Full output truncated" in result[-1]
        assert "8 lines hidden" in result[-1]
        last_line_before_trunc_msg = result[-self.LINES_IN_TRUNCATION_MSG - 1]
        assert last_line_before_trunc_msg.endswith("...")

    def test_truncates_at_4_lines_when_first_4_lines_are_GT_max_chars(self) -> None:
        expl = ["a" * 250 for x in range(10)]
        result = truncate._truncate_explanation(expl, max_lines=8, max_chars=999)
        assert result != expl
        assert len(result) == 4 + self.LINES_IN_TRUNCATION_MSG
        assert "Full output truncated" in result[-1]
        assert "7 lines hidden" in result[-1]
        last_line_before_trunc_msg = result[-self.LINES_IN_TRUNCATION_MSG - 1]
        assert last_line_before_trunc_msg.endswith("...")

    def test_truncates_at_1_line_when_first_line_is_GT_max_chars(self) -> None:
        expl = ["a" * 250 for x in range(1000)]
        result = truncate._truncate_explanation(expl, max_lines=8, max_chars=100)
        assert result != expl
        assert len(result) == 1 + self.LINES_IN_TRUNCATION_MSG
        assert "Full output truncated" in result[-1]
        assert "1000 lines hidden" in result[-1]
        last_line_before_trunc_msg = result[-self.LINES_IN_TRUNCATION_MSG - 1]
        assert last_line_before_trunc_msg.endswith("...")

    def test_full_output_truncated(self, monkeypatch, pytester: Pytester) -> None:
        """Test against full runpytest() output."""
        line_count = 7
        line_len = 100
        expected_truncated_lines = 2
        pytester.makepyfile(
            r"""
            def test_many_lines():
                a = list([str(i)[0] * %d for i in range(%d)])
                b = a[::2]
                a = '\n'.join(map(str, a))
                b = '\n'.join(map(str, b))
                assert a == b
        """
            % (line_len, line_count)
        )
        monkeypatch.delenv("CI", raising=False)

        result = pytester.runpytest()
        # without -vv, truncate the message showing a few diff lines only
        result.stdout.fnmatch_lines(
            [
                "*+ 1*",
                "*+ 3*",
                "*truncated (%d lines hidden)*use*-vv*" % expected_truncated_lines,
            ]
        )

        result = pytester.runpytest("-vv")
        result.stdout.fnmatch_lines(["* 6*"])

        monkeypatch.setenv("CI", "1")
        result = pytester.runpytest()
        result.stdout.fnmatch_lines(["* 6*"])


def test_python25_compile_issue257(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        def test_rewritten():
            assert 1 == 2
        # some comment
    """
    )
    result = pytester.runpytest()
    assert result.ret == 1
    result.stdout.fnmatch_lines(
        """
            *E*assert 1 == 2*
            *1 failed*
    """
    )


def test_rewritten(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        def test_rewritten():
            assert "@py_builtins" in globals()
    """
    )
    assert pytester.runpytest().ret == 0


def test_reprcompare_notin() -> None:
    assert callop("not in", "foo", "aaafoobbb") == [
        "'foo' not in 'aaafoobbb'",
        "",
        "'foo' is contained here:",
        "  aaafoobbb",
        "?    +++",
    ]


def test_reprcompare_whitespaces() -> None:
    assert callequal("\r\n", "\n") == [
        r"'\r\n' == '\n'",
        "",
        r"Strings contain only whitespace, escaping them using repr()",
        r"- '\n'",
        r"+ '\r\n'",
        r"?  ++",
    ]


class TestSetAssertions:
    @pytest.mark.parametrize("op", [">=", ">", "<=", "<", "=="])
    def test_set_extra_item(self, op, pytester: Pytester) -> None:
        pytester.makepyfile(
            f"""
            def test_hello():
                x = set("hello x")
                y = set("hello y")
                assert x {op} y
        """
        )

        result = pytester.runpytest()
        result.stdout.fnmatch_lines(
            [
                "*def test_hello():*",
                f"*assert x {op} y*",
            ]
        )
        if op in [">=", ">", "=="]:
            result.stdout.fnmatch_lines(
                [
                    "*E*Extra items in the right set:*",
                    "*E*'y'",
                ]
            )
        if op in ["<=", "<", "=="]:
            result.stdout.fnmatch_lines(
                [
                    "*E*Extra items in the left set:*",
                    "*E*'x'",
                ]
            )

    @pytest.mark.parametrize("op", [">", "<", "!="])
    def test_set_proper_superset_equal(self, pytester: Pytester, op) -> None:
        pytester.makepyfile(
            f"""
            def test_hello():
                x = set([1, 2, 3])
                y = x.copy()
                assert x {op} y
        """
        )

        result = pytester.runpytest()
        result.stdout.fnmatch_lines(
            [
                "*def test_hello():*",
                f"*assert x {op} y*",
                "*E*Both sets are equal*",
            ]
        )

    def test_pytest_assertrepr_compare_integration(self, pytester: Pytester) -> None:
        pytester.makepyfile(
            """
            def test_hello():
                x = set(range(100))
                y = x.copy()
                y.remove(50)
                assert x == y
        """
        )
        result = pytester.runpytest()
        result.stdout.fnmatch_lines(
            [
                "*def test_hello():*",
                "*assert x == y*",
                "*E*Extra items*left*",
                "*E*50*",
                "*= 1 failed in*",
            ]
        )


def test_assertrepr_loaded_per_dir(pytester: Pytester) -> None:
    pytester.makepyfile(test_base=["def test_base(): assert 1 == 2"])
    a = pytester.mkdir("a")
    a.joinpath("test_a.py").write_text("def test_a(): assert 1 == 2", encoding="utf-8")
    a.joinpath("conftest.py").write_text(
        'def pytest_assertrepr_compare(): return ["summary a"]', encoding="utf-8"
    )
    b = pytester.mkdir("b")
    b.joinpath("test_b.py").write_text("def test_b(): assert 1 == 2", encoding="utf-8")
    b.joinpath("conftest.py").write_text(
        'def pytest_assertrepr_compare(): return ["summary b"]', encoding="utf-8"
    )

    result = pytester.runpytest()
    result.stdout.fnmatch_lines(
        [
            "*def test_a():*",
            "*E*assert summary a*",
            "*def test_b():*",
            "*E*assert summary b*",
            "*def test_base():*",
            "*E*assert 1 == 2*",
        ]
    )


def test_assertion_options(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        def test_hello():
            x = 3
            assert x == 4
    """
    )
    result = pytester.runpytest()
    assert "3 == 4" in result.stdout.str()
    result = pytester.runpytest_subprocess("--assert=plain")
    result.stdout.no_fnmatch_line("*3 == 4*")


def test_triple_quoted_string_issue113(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        def test_hello():
            assert "" == '''
    '''"""
    )
    result = pytester.runpytest("--fulltrace")
    result.stdout.fnmatch_lines(["*1 failed*"])
    result.stdout.no_fnmatch_line("*SyntaxError*")


def test_traceback_failure(pytester: Pytester) -> None:
    p1 = pytester.makepyfile(
        """
        def g():
            return 2
        def f(x):
            assert x == g()
        def test_onefails():
            f(3)
    """
    )
    result = pytester.runpytest(p1, "--tb=long")
    result.stdout.fnmatch_lines(
        [
            "*test_traceback_failure.py F*",
            "====* FAILURES *====",
            "____*____",
            "",
            "    def test_onefails():",
            ">       f(3)",
            "",
            "*test_*.py:6: ",
            "_ _ _ *",
            # "",
            "    def f(x):",
            ">       assert x == g()",
            "E       assert 3 == 2",
            "E        +  where 2 = g()",
            "",
            "*test_traceback_failure.py:4: AssertionError",
        ]
    )

    result = pytester.runpytest(p1)  # "auto"
    result.stdout.fnmatch_lines(
        [
            "*test_traceback_failure.py F*",
            "====* FAILURES *====",
            "____*____",
            "",
            "    def test_onefails():",
            ">       f(3)",
            "",
            "*test_*.py:6: ",
            "",
            "    def f(x):",
            ">       assert x == g()",
            "E       assert 3 == 2",
            "E        +  where 2 = g()",
            "",
            "*test_traceback_failure.py:4: AssertionError",
        ]
    )


def test_exception_handling_no_traceback(pytester: Pytester) -> None:
    """Handle chain exceptions in tasks submitted by the multiprocess module (#1984)."""
    p1 = pytester.makepyfile(
        """
        from multiprocessing import Pool

        def process_task(n):
            assert n == 10

        def multitask_job():
            tasks = [1]
            with Pool(processes=1) as pool:
                pool.map(process_task, tasks)

        def test_multitask_job():
            multitask_job()
    """
    )
    pytester.syspathinsert()
    result = pytester.runpytest(p1, "--tb=long")
    result.stdout.fnmatch_lines(
        [
            "====* FAILURES *====",
            "*multiprocessing.pool.RemoteTraceback:*",
            "Traceback (most recent call last):",
            "*assert n == 10",
            "The above exception was the direct cause of the following exception:",
            "> * multitask_job()",
        ]
    )


@pytest.mark.skipif("'__pypy__' in sys.builtin_module_names")
@pytest.mark.parametrize(
    "cmdline_args, warning_output",
    [
        (
            ["-OO", "-m", "pytest", "-h"],
            ["warning :*PytestConfigWarning:*assert statements are not executed*"],
        ),
        (
            ["-OO", "-m", "pytest"],
            [
                "=*= warnings summary =*=",
                "*PytestConfigWarning:*assert statements are not executed*",
            ],
        ),
        (
            ["-OO", "-m", "pytest", "--assert=plain"],
            [
                "=*= warnings summary =*=",
                "*PytestConfigWarning: ASSERTIONS ARE NOT EXECUTED and FAILING TESTS WILL PASS.  "
                "Are you using python -O?",
            ],
        ),
    ],
)
def test_warn_missing(pytester: Pytester, cmdline_args, warning_output) -> None:
    pytester.makepyfile("")

    result = pytester.run(sys.executable, *cmdline_args)
    result.stdout.fnmatch_lines(warning_output)


def test_recursion_source_decode(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        def test_something():
            pass
    """
    )
    pytester.makeini(
        """
        [pytest]
        python_files = *.py
    """
    )
    result = pytester.runpytest("--collect-only")
    result.stdout.fnmatch_lines(
        [
            "  <Module*>",
        ]
    )


def test_AssertionError_message(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        def test_hello():
            x,y = 1,2
            assert 0, (x,y)
    """
    )
    result = pytester.runpytest()
    result.stdout.fnmatch_lines(
        """
        *def test_hello*
        *assert 0, (x,y)*
        *AssertionError: (1, 2)*
    """
    )


def test_diff_newline_at_end(pytester: Pytester) -> None:
    pytester.makepyfile(
        r"""
        def test_diff():
            assert 'asdf' == 'asdf\n'
    """
    )

    result = pytester.runpytest()
    result.stdout.fnmatch_lines(
        r"""
        *assert 'asdf' == 'asdf\n'
        *  - asdf
        *  ?     -
        *  + asdf
    """
    )


@pytest.mark.filterwarnings("default")
def test_assert_tuple_warning(pytester: Pytester) -> None:
    msg = "assertion is always true"
    pytester.makepyfile(
        """
        def test_tuple():
            assert(False, 'you shall not pass')
    """
    )
    result = pytester.runpytest()
    result.stdout.fnmatch_lines([f"*test_assert_tuple_warning.py:2:*{msg}*"])

    # tuples with size != 2 should not trigger the warning
    pytester.makepyfile(
        """
        def test_tuple():
            assert ()
    """
    )
    result = pytester.runpytest()
    assert msg not in result.stdout.str()


def test_assert_indirect_tuple_no_warning(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        def test_tuple():
            tpl = ('foo', 'bar')
            assert tpl
    """
    )
    result = pytester.runpytest()
    output = "\n".join(result.stdout.lines)
    assert "WR1" not in output


def test_assert_with_unicode(pytester: Pytester) -> None:
    pytester.makepyfile(
        """\
        def test_unicode():
            assert '유니코드' == 'Unicode'
        """
    )
    result = pytester.runpytest()
    result.stdout.fnmatch_lines(["*AssertionError*"])


def test_raise_unprintable_assertion_error(pytester: Pytester) -> None:
    pytester.makepyfile(
        r"""
        def test_raise_assertion_error():
            raise AssertionError('\xff')
    """
    )
    result = pytester.runpytest()
    result.stdout.fnmatch_lines(
        [r">       raise AssertionError('\xff')", "E       AssertionError: *"]
    )


def test_raise_assertion_error_raising_repr(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        class RaisingRepr(object):
            def __repr__(self):
                raise Exception()
        def test_raising_repr():
            raise AssertionError(RaisingRepr())
    """
    )
    result = pytester.runpytest()
    result.stdout.fnmatch_lines(["E       AssertionError: <exception str() failed>"])


def test_issue_1944(pytester: Pytester) -> None:
    pytester.makepyfile(
        """
        def f():
            return

        assert f() == 10
    """
    )
    result = pytester.runpytest()
    result.stdout.fnmatch_lines(["*1 error*"])
    assert (
        "AttributeError: 'Module' object has no attribute '_obj'"
        not in result.stdout.str()
    )


def test_exit_from_assertrepr_compare(monkeypatch) -> None:
    def raise_exit(obj):
        outcomes.exit("Quitting debugger")

    monkeypatch.setattr(util, "istext", raise_exit)

    with pytest.raises(outcomes.Exit, match="Quitting debugger"):
        callequal(1, 1)


def test_assertion_location_with_coverage(pytester: Pytester) -> None:
    """This used to report the wrong location when run with coverage (#5754)."""
    p = pytester.makepyfile(
        """
        def test():
            assert False, 1
            assert False, 2
        """
    )
    result = pytester.runpytest(str(p))
    result.stdout.fnmatch_lines(
        [
            ">       assert False, 1",
            "E       AssertionError: 1",
            "E       assert False",
            "*= 1 failed in*",
        ]
    )


def test_reprcompare_verbose_long() -> None:
    a = {f"v{i}": i for i in range(11)}
    b = a.copy()
    b["v2"] += 10
    lines = callop("==", a, b, verbose=2)
    assert lines is not None
    assert lines[0] == (
        "{'v0': 0, 'v1': 1, 'v2': 2, 'v3': 3, 'v4': 4, 'v5': 5, "
        "'v6': 6, 'v7': 7, 'v8': 8, 'v9': 9, 'v10': 10}"
        " == "
        "{'v0': 0, 'v1': 1, 'v2': 12, 'v3': 3, 'v4': 4, 'v5': 5, "
        "'v6': 6, 'v7': 7, 'v8': 8, 'v9': 9, 'v10': 10}"
    )


@pytest.mark.parametrize("enable_colors", [True, False])
@pytest.mark.parametrize(
    ("test_code", "expected_lines"),
    (
        (
            """
            def test():
                assert [0, 1] == [0, 2]
            """,
            [
                "{bold}{red}E         At index 1 diff: {reset}{number}1{hl-reset}{endline} != {reset}{number}2*",
                "{bold}{red}E         {light-red}-     2,{hl-reset}{endline}{reset}",
                "{bold}{red}E         {light-green}+     1,{hl-reset}{endline}{reset}",
            ],
        ),
        (
            """
            def test():
                assert {f"number-is-{i}": i for i in range(1, 6)} == {
                    f"number-is-{i}": i for i in range(5)
                }
            """,
            [
                "{bold}{red}E         Common items:{reset}",
                "{bold}{red}E         {reset}{{{str}'{hl-reset}{str}number-is-1{hl-reset}{str}'{hl-reset}: {number}1*",
                "{bold}{red}E         Left contains 1 more item:{reset}",
                "{bold}{red}E         {reset}{{{str}'{hl-reset}{str}number-is-5{hl-reset}{str}'{hl-reset}: {number}5*",
                "{bold}{red}E         Right contains 1 more item:{reset}",
                "{bold}{red}E         {reset}{{{str}'{hl-reset}{str}number-is-0{hl-reset}{str}'{hl-reset}: {number}0*",
                "{bold}{red}E         {reset}{light-gray} {hl-reset} {{{endline}{reset}",
                "{bold}{red}E         {light-gray} {hl-reset}     'number-is-1': 1,{endline}{reset}",
                "{bold}{red}E         {light-green}+     'number-is-5': 5,{hl-reset}{endline}{reset}",
            ],
        ),
    ),
)
def test_comparisons_handle_colors(
    pytester: Pytester, color_mapping, enable_colors, test_code, expected_lines
) -> None:
    p = pytester.makepyfile(test_code)
    result = pytester.runpytest(
        f"--color={'yes' if enable_colors else 'no'}", "-vv", str(p)
    )
    formatter = (
        color_mapping.format_for_fnmatch
        if enable_colors
        else color_mapping.strip_colors
    )

    result.stdout.fnmatch_lines(formatter(expected_lines), consecutive=False)


def test_fine_grained_assertion_verbosity(pytester: Pytester):
    long_text = "Lorem ipsum dolor sit amet " * 10
    p = pytester.makepyfile(
        f"""
        def test_ok():
            pass


        def test_words_fail():
            fruits1 = ["banana", "apple", "grapes", "melon", "kiwi"]
            fruits2 = ["banana", "apple", "orange", "melon", "kiwi"]
            assert fruits1 == fruits2


        def test_numbers_fail():
            number_to_text1 = {{str(x): x for x in range(5)}}
            number_to_text2 = {{str(x * 10): x * 10 for x in range(5)}}
            assert number_to_text1 == number_to_text2


        def test_long_text_fail():
            long_text = "{long_text}"
            assert "hello world" in long_text
        """
    )
    pytester.makeini(
        """
        [pytest]
        verbosity_assertions = 2
        """
    )
    result = pytester.runpytest(p)

    result.stdout.fnmatch_lines(
        [
            f"{p.name} .FFF                            [100%]",
            "E         At index 2 diff: 'grapes' != 'orange'",
            "E         Full diff:",
            "E           [",
            "E               'banana',",
            "E               'apple',",
            "E         -     'orange',",
            "E         ?      ^  ^^",
            "E         +     'grapes',",
            "E         ?      ^  ^ +",
            "E               'melon',",
            "E               'kiwi',",
            "E           ]",
            "E         Full diff:",
            "E           {",
            "E               '0': 0,",
            "E         -     '10': 10,",
            "E         ?       -    -",
            "E         +     '1': 1,",
            "E         -     '20': 20,",
            "E         ?       -    -",
            "E         +     '2': 2,",
            "E         -     '30': 30,",
            "E         ?       -    -",
            "E         +     '3': 3,",
            "E         -     '40': 40,",
            "E         ?       -    -",
            "E         +     '4': 4,",
            "E           }",
            f"E       AssertionError: assert 'hello world' in '{long_text}'",
        ]
    )
